""" Testing Image spaces
"""

import numpy as np

import nibabel as nib
from nibabel.affines import from_matvec

from ..image import Image, rollimg
from ..image_spaces import (is_xyz_affable, as_xyz_image, xyz_affine,
                            make_xyz_image)
from ...reference.coordinate_system import CoordinateSystem as CS
from ...reference.coordinate_map import AffineTransform
from ...reference.spaces import (vox2mni, vox2talairach, voxel_csm,
                                 mni_space, talairach_space,
                                 AffineError, AxesError,
                                 XYZSpace, SpaceError)

from numpy.testing import (assert_array_almost_equal, assert_array_equal)

from nose.tools import assert_true, assert_false, assert_equal, assert_raises

def test_image_xyz_affine():
    # Test getting the image xyz affines
    arr = np.arange(24).reshape((2,3,4))
    aff = np.diag([2,3,4,1])
    img = Image(arr, vox2mni(aff))
    assert_true(is_xyz_affable(img))
    assert_array_equal(xyz_affine(img), aff)
    arr4 = np.arange(24).reshape((1,2,3,4))
    img4 = Image(arr4, vox2mni(np.diag([2,3,4,5,1])))
    assert_true(is_xyz_affable(img4))
    img4_r = img4.reordered_axes([3,2,0,1])
    assert_false(is_xyz_affable(img4_r))
    assert_raises(AxesError, xyz_affine, img4_r)
    nimg = nib.Nifti1Image(arr, aff)
    assert_true(is_xyz_affable(nimg))
    assert_array_equal(xyz_affine(nimg), aff)
    # Any dimensions not spatial, AxesError
    d_cs = CS('ijk', 'voxels')
    r_cs = CS(('mni-x=L->R', 'mni-y=P->A', 'mni-q'), 'mni')
    cmap = AffineTransform(d_cs,r_cs, aff)
    img = Image(arr, cmap)
    assert_raises(AxesError, xyz_affine, img)
    # Can pass in own validator
    my_valtor = dict(blind='x', leading='y', ditch='z')
    r_cs = CS(('blind', 'leading', 'ditch'), 'fall')
    cmap = AffineTransform(d_cs, r_cs, aff)
    img = Image(arr, cmap)
    assert_raises(AxesError, xyz_affine, img)
    assert_array_equal(xyz_affine(img, my_valtor), aff)


def test_image_as_xyz_image():
    # Test getting xyz affable version of the image
    arr = np.arange(24).reshape((1,2,3,4))
    aff = np.diag([2,3,4,5,1])
    img = Image(arr, vox2mni(aff))
    img_r = as_xyz_image(img)
    assert_true(img is img_r)
    # Reorder, reverse reordering, test != and ==
    for order in ((3, 0, 1, 2), (0, 3, 1, 2)):
        img_ro_out = img.reordered_reference(order)
        img_ro_in = img.reordered_axes(order)
        img_ro_both = img_ro_out.reordered_axes(order)
        for tmap in (img_ro_out, img_ro_in, img_ro_both):
            assert_false(is_xyz_affable(tmap))
            img_r = as_xyz_image(tmap)
            assert_false(tmap is img_r)
            assert_equal(img, img_r)
            assert_array_equal(img.get_data(), img_r.get_data())
    img_t0 = rollimg(img, 't')
    assert_false(is_xyz_affable(img_t0))
    img_t0_r = as_xyz_image(img_t0)
    assert_false(img_t0 is img_t0_r)
    assert_array_equal(img.get_data(), img_t0_r.get_data())
    assert_equal(img.coordmap, img_t0_r.coordmap)
    # Test against nibabel image
    nimg = nib.Nifti1Image(arr, np.diag([2,3,4,1]))
    nimg_r = as_xyz_image(nimg)
    assert_true(nimg is nimg_r)
    # It's sometimes impossible to make an xyz affable image
    # If the xyz coordinates depend on the time coordinate
    aff = np.array([[2, 0, 0, 2, 20],
                    [0, 3, 0, 0, 21],
                    [0, 0, 4, 0, 22],
                    [0, 0, 0, 5, 23],
                    [0, 0, 0, 0, 1]])
    img = Image(arr, vox2mni(aff))
    assert_raises(AffineError, as_xyz_image, img)
    # If any dimensions not spatial, AxesError
    arr = np.arange(24).reshape((2,3,4))
    aff = np.diag([2,3,4,1])
    d_cs = CS('ijk', 'voxels')
    r_cs = CS(('mni-x=L->R', 'mni-y=P->A', 'mni-q'), 'mni')
    cmap = AffineTransform(d_cs, r_cs, aff)
    img = Image(arr, cmap)
    assert_raises(AxesError, as_xyz_image, img)
    # Can pass in own validator
    my_valtor = dict(blind='x', leading='y', ditch='z')
    r_cs = CS(('blind', 'leading', 'ditch'), 'fall')
    cmap = AffineTransform(d_cs, r_cs, aff)
    img = Image(arr, cmap)
    assert_raises(AxesError, as_xyz_image, img)
    assert_true(as_xyz_image(img, my_valtor) is img)


def test_image_xyza_slices():
    # Jonathan found some nastiness where xyz present in output but there was
    # not corresponding axis for x in the input
    arr = np.arange(24).reshape((1,2,3,4))
    aff = np.diag([2,3,4,5,1])
    img = Image(arr, vox2mni(aff))
    img0 = img[0] # slice in X
    # The result does not have an input axis corresponding to x, and should
    # raise an error
    assert_raises(AxesError, as_xyz_image, img0)
    img0r = img0.reordered_reference([1,0,2,3]).reordered_axes([2,0,1])
    assert_raises(AxesError, as_xyz_image, img0r)


def test_make_xyz_image():
    # Standard neuro image creator
    arr = np.arange(24).reshape((1,2,3,4))
    aff = np.diag([2,3,4,1])
    img = make_xyz_image(arr, aff, 'mni')
    assert_equal(img.coordmap, vox2mni(aff, 1.0))
    assert_array_equal(img.get_data(), arr)
    img = make_xyz_image(arr, aff, 'talairach')
    assert_equal(img.coordmap, vox2talairach(aff, 1.0))
    assert_array_equal(img.get_data(), arr)
    img = make_xyz_image(arr, aff, talairach_space)
    assert_equal(img.coordmap, vox2talairach(aff, 1.0))
    # Unknown space as string raises SpaceError
    assert_raises(SpaceError, make_xyz_image, arr, aff, 'unlikely space name')
    funky_space = XYZSpace('hija')
    img = make_xyz_image(arr, aff, funky_space)
    csm = funky_space.to_coordsys_maker('t')
    in_cs = CS('ijkl', 'voxels')
    exp_cmap = AffineTransform(in_cs, csm(4), np.diag([2, 3, 4, 1, 1]))
    assert_equal(img.coordmap, exp_cmap)
    # Affine must be 4, 4
    bad_aff = np.diag([2,3,4,2,1])
    assert_raises(ValueError, make_xyz_image, arr, bad_aff, 'mni')
    # Can pass added zooms
    img = make_xyz_image(arr, (aff, (2.,)), 'mni')
    assert_equal(img.coordmap, vox2mni(aff, 2.0))
    # Also as scalar
    img = make_xyz_image(arr, (aff, 2.), 'mni')
    assert_equal(img.coordmap, vox2mni(aff, 2.0))
    # Must match length of needed zooms
    arr5 = arr[...,None]
    assert_raises(ValueError, make_xyz_image, arr5, (aff, 2.), 'mni')
    img = make_xyz_image(arr5, (aff, (2., 3.)), 'mni')
    assert_equal(img.coordmap, vox2mni(aff, (2.0, 3.0)))
    # Always xyz affable after creation
    assert_array_equal(xyz_affine(img), aff)
    assert_true(is_xyz_affable(img))
    # Need at least 3 dimensions in data
    assert_raises(ValueError, make_xyz_image, np.zeros((2,3)), aff, 'mni')
    # Check affines don't round / floor floating point
    aff = np.diag([2.1, 3, 4, 1])
    img = make_xyz_image(np.zeros((2, 3, 4)), aff, 'scanner')
    assert_array_equal(img.coordmap.affine, aff)
    img = make_xyz_image(np.zeros((2, 3, 4, 5)), aff, 'scanner')
    assert_array_equal(img.coordmap.affine, np.diag([2.1, 3, 4, 1, 1]))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from nose.tools import assert_equal, assert_raises

from numpy.testing import assert_array_almost_equal, assert_array_equal
import numpy as np

from .... import load_image
from ....testing import funcfile
from ....fixes.nibabel import io_orientation

from ..groupwise_registration import Image4d, resample4d, FmriRealign4d
from ..affine import Rigid

im = load_image(funcfile)


def test_scanner_time():
    im4d = Image4d(im.get_data(), im.affine, tr=2.,
                   slice_order='ascending', interleaved=False)
    assert_equal(im4d.scanner_time(0, 0), 0.)
    assert_equal(im4d.scanner_time(0, im4d.tr), 1.)
    assert_equal(im4d.scanner_time(1, im4d.tr_slices), 0.)


def test_slice_info():
    im4d = Image4d(im.get_data(), im.affine, tr=2.,
                   slice_info=(1, -1))
    assert_equal(im4d.slice_axis, 1)
    assert_equal(im4d.slice_direction, -1)


def _test_image4d_init(nslices):
    data = np.zeros((3, 4, nslices, 6))
    aff = np.eye(4)
    tr = 2.0
    img4d = Image4d(data, aff, tr)
    assert_array_equal(img4d.slice_order, range(nslices))
    img4d = Image4d(data, aff, tr, slice_order='ascending')
    assert_array_equal(img4d.slice_order, range(nslices))
    img4d = Image4d(data, aff, tr, slice_order='descending')
    assert_array_equal(img4d.slice_order, range(nslices)[::-1])
    # test interleaved slice order
    slice_order = range(nslices)[::2] + range(nslices)[1::2]
    img4d = Image4d(data, aff, tr, slice_order='ascending', interleaved=True)
    assert_array_equal(img4d.slice_order, slice_order)
    slice_order.reverse()
    img4d = Image4d(data, aff, tr, slice_order='descending', interleaved=True)
    assert_array_equal(img4d.slice_order, slice_order)
    # can pass array
    img4d = Image4d(data, aff, tr, slice_order=np.arange(nslices))
    assert_array_equal(img4d.slice_order, range(nslices))
    # or list
    img4d = Image4d(data, aff, tr, slice_order=range(nslices))
    assert_array_equal(img4d.slice_order, range(nslices))
    # but raises exception in case of the incorrect slice indexes
    for bad_slice_order in (
        [0],                     # insufficient
        np.arange(nslices)-1,    # negative etc
        np.arange(nslices) + 0.1, # floats
        range(nslices//2)*2,     # twice the same (would match in length for even nslices)
        ):
        assert_raises(ValueError, Image4d, data, aff, tr, slice_order=bad_slice_order)


def test_image4d_init_5slices():
    _test_image4d_init(5)


def test_image4d_init_6slices():
    _test_image4d_init(6)


def test_slice_timing():
    affine = np.eye(4)
    affine[0:3, 0:3] = im.affine[0:3, 0:3]
    im4d = Image4d(im.get_data(), affine, tr=2., tr_slices=0.0)
    x = resample4d(im4d, [Rigid() for i in range(im.shape[3])])
    assert_array_almost_equal(im4d.get_data(), x)


def test_realign4d_no_time_interp():
    runs = [im, im]
    R = FmriRealign4d(runs, slice_order=None, time_interp=False)


def test_realign4d():
    """
    This tests whether realign4d yields the same results depending on
    whether the slice order is input explicitely or as
    slice_order='ascending'.
    
    Due to the very small size of the image used for testing (only 3
    slices), optimization is numerically unstable. It seems to make
    the default optimizer, namely scipy.fmin.fmin_ncg, adopt a random
    behavior. To work around the resulting inconsistency in results,
    we use nipy.optimize.fmin_steepest as the optimizer, although it's
    generally not recommended in practice.
    """
    runs = [im, im]
    orient = io_orientation(im.affine)
    slice_axis = int(np.where(orient[:, 0] == 2)[0])
    R1 = FmriRealign4d(runs, tr=2., slice_order='ascending')
    R1.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    R2 = FmriRealign4d(runs, tr=2., slice_order=range(im.shape[slice_axis]))
    R2.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    for r in range(2):
        for i in range(im.shape[3]):
            assert_array_almost_equal(R1._transforms[r][i].translation,
                                      R2._transforms[r][i].translation)
            assert_array_almost_equal(R1._transforms[r][i].rotation,
                                      R2._transforms[r][i].rotation)
    for i in range(im.shape[3]):
            assert_array_almost_equal(R1._mean_transforms[r].translation,
                                      R2._mean_transforms[r].translation)
            assert_array_almost_equal(R1._mean_transforms[r].rotation,
                                      R2._mean_transforms[r].rotation)
    



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
import scipy.stats

TINY = 1e-15


def zscore(pvalue):
    """ Return the z-score corresponding to a given p-value.
    """
    pvalue = np.minimum(np.maximum(pvalue, TINY), 1. - TINY)
    z = scipy.stats.norm.isf(pvalue)
    return z



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Bayesian Gaussian Mixture Model Classes:
contains the basic fields and methods of Bayesian GMMs
the high level functions are/should be binded in C

The base class BGMM relies on an implementation that perfoms Gibbs sampling

A derived class VBGMM uses Variational Bayes inference instead

A third class is introduces to take advnatge of the old C-bindings,
but it is limited to diagonal covariance models

Author : Bertrand Thirion, 2008-2011
"""
import numpy as np
import numpy.random as nr
from scipy.linalg import inv, cholesky, eigvalsh
from scipy.special import gammaln
import math

from .utils import kmeans
from gmm import GMM

##################################################################
# ancillary functions ############################################
##################################################################


def detsh(H):
    """
    Routine for the computation of determinants of symmetric positive
    matrices

    Parameters
    ----------
    H array of shape(n,n)
      the input matrix, assumed symmmetric and positive

    Returns
    -------
    dh: float, the determinant
    """
    return np.prod(eigvalsh(H))


def dirichlet_eval(w, alpha):
    """
    Evaluate the probability of a certain discrete draw w
    from the Dirichlet density with parameters alpha

    Parameters
    ----------
    w: array of shape (n)
    alpha: array of shape (n)
    """
    if np.shape(w) != np.shape(alpha):
        raise ValueError("incompatible dimensions")
    loge = np.sum((alpha-1) * np.log(w))
    logb = np.sum(gammaln(alpha)) - gammaln(alpha.sum())
    loge -= logb
    return np.exp(loge)


def generate_normals(m, P):
    """ Generate a Gaussian sample with mean m and precision P

    Parameters
    ----------
    m array of shape n: the mean vector
    P array of shape (n,n): the precision matrix

    Returns
    -------
    ng : array of shape(n): a draw from the gaussian density
    """
    icp = inv(cholesky(P))
    ng = nr.randn(m.shape[0])
    ng = np.dot(ng, icp)
    ng += m
    return ng


def generate_Wishart(n, V):
    """
    Generate a sample from Wishart density

    Parameters
    ----------
    n: float,
        the number of degrees of freedom of the Wishart density
    V: array of shape (n,n)
       the scale matrix of the Wishart density

    Returns
    -------
    W: array of shape (n,n)
       the draw from Wishart density
    """
    icv = cholesky(V)
    p = V.shape[0]
    A = nr.randn(p, p)
    for i in range(p):
        A[i, i:] = 0
        A[i, i] = np.sqrt(nr.chisquare(n - i))
    R = np.dot(icv, A)
    W = np.dot(R, R.T)
    return W


def wishart_eval(n, V, W, dV=None, dW=None, piV=None):
    """Evaluation of the  probability of W under Wishart(n,V)

    Parameters
    ----------
    n: float,
        the number of degrees of freedom (dofs)
    V: array of shape (n,n)
        the scale matrix of the Wishart density
    W: array of shape (n,n)
        the sample to be evaluated
    dV: float, optional,
        determinant of V
    dW: float, optional,
        determinant of W
    piV: array of shape (n,n), optional
         inverse of V

    Returns
    -------
    (float) the density
    """
    # check that shape(V)==shape(W)
    p = V.shape[0]
    if dV == None:
        dV = detsh(V)
    if dW == None:
        dW = detsh(W)
    if piV == None:
        piV = inv(V)
    ldW = math.log(dW) * (n - p - 1) / 2
    ltr = - np.trace(np.dot(piV, W)) / 2
    la = (n * p * math.log(2) + math.log(dV) * n) / 2
    lg = math.log(math.pi) * p * (p - 1) / 4
    lg += gammaln(np.arange(n - p + 1, n + 1).astype(np.float) / 2).sum()
    lt = ldW + ltr - la - lg
    return math.exp(lt)


def normal_eval(mu, P, x, dP=None):
    """ Probability of x under normal(mu, inv(P))

    Parameters
    ----------
    mu: array of shape (n),
        the mean parameter
    P: array of shape (n, n),
       the precision matrix
    x: array of shape (n),
       the data to be evaluated

    Returns
    -------
    (float) the density
    """
    dim = P.shape[0]
    if dP == None:
        dP = detsh(P)

    w0 = math.log(dP) - dim * math.log(2 * math.pi)
    w0 /= 2
    dx = mu - x
    q = np.dot(np.dot(P, dx), dx)
    w = w0 - q / 2
    like = math.exp(w)
    return like


def generate_perm(k, nperm=100):
    """
    returns an array of shape(nbperm, k) representing
    the permutations of k elements

    Parameters
    ----------
    k, int the number of elements to be permuted
    nperm=100 the maximal number of permutations
    if gamma(k+1)>nperm: only nperm random draws are generated

    Returns
    -------
    p: array of shape(nperm,k): each row is permutation of k
    """
    from scipy.special import gamma
    if k == 1:
        return np.reshape(np.array([0]), (1, 1)).astype(np.int)
    if gamma(k + 1) < nperm:
        # exhaustive permutations
        aux = generate_perm(k - 1)
        n = aux.shape[0]
        perm = np.zeros((n * k, k)).astype(np.int)
        for i in range(k):
            perm[i * n:(i + 1) * n, :i] = aux[:, :i]
            perm[i * n:(i + 1) * n, i] = k-1
            perm[i * n:(i + 1) * n, i + 1:] = aux[:, i:]
    else:
        from numpy.random import rand
        perm = np.zeros((nperm, k)).astype(np.int)
        for i in range(nperm):
            p = np.argsort(rand(k))
            perm[i] = p
    return perm


def multinomial(probabilities):
    """
    Generate samples form a miltivariate distribution

    Parameters
    ----------
    probabilities: array of shape (nelements, nclasses):
                likelihood of each element belongin to each class
                each row is assumedt to sum to 1
                One sample is draw from each row, resulting in

    Returns
    -------
    z array of shape (nelements): the draws,
      that take values in [0..nclasses-1]
    """
    nvox = probabilities.shape[0]
    nclasses = probabilities.shape[1]
    cuml = np.zeros((nvox, nclasses + 1))
    cuml[:, 1:] = np.cumsum(probabilities, 1)
    aux = np.random.rand(nvox, 1)
    z = np.argmax(aux < cuml, 1)-1
    return z


def dkl_gaussian(m1, P1, m2, P2):
    """
    Returns the KL divergence between gausians densities

    Parameters
    ----------
    m1: array of shape (n),
        the mean parameter of the first density
    P1: array of shape(n,n),
        the precision parameters of the first density
    m2: array of shape (n),
        the mean parameter of the second density
    P2: array of shape(n,n),
        the precision parameters of the second density
    """
    tiny = 1.e-15
    dim = np.size(m1)
    if m1.shape != m2.shape:
        raise ValueError("incompatible dimensions for m1 and m2")
    if P1.shape != P2.shape:
        raise ValueError("incompatible dimensions for P1 and P2")
    if P1.shape[0] != dim:
        raise ValueError("incompatible dimensions for m1 and P1")

    d1 = max(detsh(P1), tiny)
    d2 = max(detsh(P2), tiny)
    dkl = np.log(d1 / d2) + np.trace(np.dot(P2, inv(P1))) - dim
    dkl += np.dot(np.dot((m1 - m2).T, P2), (m1 - m2))
    dkl /= 2
    return dkl


def dkl_wishart(a1, B1, a2, B2):
    """
    returns the KL divergence bteween two Wishart distribution of
    parameters (a1,B1) and (a2,B2),

    Parameters
    ----------
    a1: Float,
        degrees of freedom of the first density
    B1: array of shape(n,n),
        scale matrix of the  first density
    a2: Float,
        degrees of freedom of the second density
    B2: array of shape(n,n),
        scale matrix of the second density

    Returns
    -------
    dkl: float, the Kullback-Leibler divergence
    """
    from scipy.special import psi, gammaln
    tiny = 1.e-15
    if B1.shape != B2.shape:
        raise ValueError("incompatible dimensions for B1 and B2")

    dim = B1.shape[0]
    d1 = max(detsh(B1), tiny)
    d2 = max(detsh(B2), tiny)
    lgc = dim * (dim - 1) * math.log(np.pi) / 4
    lg1 = lgc
    lg2 = lgc
    lw1 = - math.log(d1) + dim * math.log(2)
    lw2 = - math.log(d2) + dim * math.log(2)
    for i in range(dim):
        lg1 += gammaln((a1 - i) / 2)
        lg2 += gammaln((a2 - i) / 2)
        lw1 += psi((a1 - i) / 2)
        lw2 += psi((a2 - i) / 2)
    lz1 = 0.5 * a1 * dim * math.log(2) - 0.5 * a1 * math.log(d1) + lg1
    lz2 = 0.5 * a2 * dim * math.log(2) - 0.5 * a2 * math.log(d2) + lg2
    dkl = (a1 - dim - 1) * lw1 - (a2 - dim - 1) * lw2 - a1 * dim
    dkl += a1 * np.trace(np.dot(B2, inv(B1)))
    dkl /= 2
    dkl += (lz2 - lz1)
    return dkl


def dkl_dirichlet(w1, w2):
    """ Returns the KL divergence between two dirichlet distribution

    Parameters
    ----------
    w1: array of shape(n),
        the parameters of the first dirichlet density
    w2: array of shape(n),
        the parameters of the second dirichlet density
    """
    if w1.shape != w2.shape:
        raise ValueError("incompatible dimensions for w1 and w2")

    dkl = 0
    from scipy.special import gammaln, psi
    dkl = np.sum(gammaln(w2)) - np.sum(gammaln(w1))
    dkl += gammaln(np.sum(w1)) - gammaln(np.sum(w2))
    dkl += np.sum((w1 - w2) * (psi(w1) - psi(np.sum(w1))))
    return dkl


#######################################################################
#  main GMM class #####################################################
#######################################################################


class BGMM(GMM):
    """
    This class implements Bayesian GMMs

    this class contains the follwing fields
    k: int,
       the number of components in the mixture
    dim: int,
         the dimension of the data
    means: array of shape (k, dim)
           all the means of the components
    precisions: array of shape (k, dim, dim)
                the precisions of the componenets
    weights: array of shape (k):
             weights of the mixture
    shrinkage: array of shape (k):
               scaling factor of the posterior precisions on the mean
    dof: array of shape (k)
         the degrees of freedom of the components
    prior_means: array of shape (k, dim):
                 the prior on the components means
    prior_scale: array of shape (k, dim):
                 the prior on the components precisions
    prior_dof: array of shape (k):
                the prior on the dof (should be at least equal to dim)
    prior_shrinkage: array of shape (k):
                     scaling factor of the prior precisions on the mean
    prior_weights: array of shape (k)
                   the prior on the components weights
    shrinkage: array of shape (k):
               scaling factor of the posterior precisions on the mean
    dof : array of shape (k): the posterior dofs

    fixme
    -----
    only 'full' precision is supported
    """

    def __init__(self, k=1, dim=1, means=None, precisions=None,
                 weights=None, shrinkage=None, dof=None):
        """
        Initialize the structure with the dimensions of the problem
        Eventually provide different terms
        """
        GMM.__init__(self, k, dim, 'full', means, precisions, weights)
        self.shrinkage = shrinkage
        self.dof = dof

        if self.shrinkage == None:
            self.shrinkage = np.ones(self.k)

        if self.dof == None:
            self.dof = np.ones(self.k)

        if self.precisions != None:
            self._detp = [detsh(self.precisions[k]) for k in range(self.k)]

    def check(self):
        """
        Checking the shape of sifferent matrices involved in the model
        """
        GMM.check(self)

        if self.prior_means.shape[0] != self.k:
            raise ValueError("Incorrect dimension for self.prior_means")
        if self.prior_means.shape[1] != self.dim:
            raise ValueError("Incorrect dimension for self.prior_means")
        if self.prior_scale.shape[0] != self.k:
            raise ValueError("Incorrect dimension for self.prior_scale")
        if self.prior_scale.shape[1] != self.dim:
            raise ValueError("Incorrect dimension for self.prior_scale")
        if self.prior_dof.shape[0] != self.k:
            raise ValueError("Incorrect dimension for self.prior_dof")
        if self.prior_weights.shape[0] != self.k:
            raise ValueError("Incorrect dimension for self.prior_weights")

    def set_priors(self, prior_means, prior_weights, prior_scale, prior_dof,
                   prior_shrinkage):
        """
        Set the prior of the BGMM

        Parameters
        ----------
        prior_means: array of shape (self.k,self.dim)
        prior_weights: array of shape (self.k)
        prior_scale: array of shape (self.k,self.dim,self.dim)
        prior_dof: array of shape (self.k)
        prior_shrinkage: array of shape (self.k)
        """
        self.prior_means = prior_means
        self.prior_weights = prior_weights
        self.prior_scale = prior_scale
        self.prior_dof = prior_dof
        self.prior_shrinkage = prior_shrinkage

        # cache some pre-computations
        self._dets = [detsh(self.prior_scale[k]) for k in range(self.k)]
        self._inv_prior_scale = np.array([inv(self.prior_scale[k])
                                          for k in range(self.k)])

        self.check()

    def guess_priors(self, x, nocheck=0):
        """
        Set the priors in order of having them weakly uninformative
        this is from  Fraley and raftery;
        Journal of Classification 24:155-181 (2007)

        Parameters
        ----------
        x, array of shape (nb_samples,self.dim)
           the data used in the estimation process
        nocheck: boolean, optional,
                 if nocheck==True, check is skipped
        """
        # a few parameters
        small = 0.01
        elshape = (1, self.dim, self.dim)
        mx = np.reshape(x.mean(0), (1, self.dim))
        dx = x - mx
        vx = np.dot(dx.T, dx) / x.shape[0]
        px = np.reshape(np.diag(1.0 / np.diag(vx)), elshape)
        px *= np.exp(2.0 / self.dim * math.log(self.k))

        # set the priors
        self.prior_means = np.repeat(mx, self.k, 0)
        self.prior_weights = np.ones(self.k)
        self.prior_scale = np.repeat(px, self.k, 0)
        self.prior_dof = np.ones(self.k) * (self.dim + 2)
        self.prior_shrinkage = np.ones(self.k) * small

        # cache some pre-computations
        self._dets = np.ones(self.k) * detsh(px[0])
        self._inv_prior_scale = np.repeat(
            np.reshape(inv(px[0]), elshape), self.k, 0)

        # check that everything is OK
        if nocheck == True:
            self.check()

    def initialize(self, x):
        """
        initialize z using a k-means algorithm, then upate the parameters

        Parameters
        ----------
        x: array of shape (nb_samples,self.dim)
           the data used in the estimation process
        """
        if self.k > 1:
            cent, z, J = kmeans(x, self.k)
        else:
            z = np.zeros(x.shape[0]).astype(np.int)
        self.update(x, z)

    def pop(self, z):
        """
        compute the population, i.e. the statistics of allocation

        Parameters
        ----------
        z array of shape (nb_samples), type = np.int
          the allocation variable

        Returns
        -------
        hist : array shape (self.k) count variable
        """
        hist = np.array([np.sum(z == k) for k in range(self.k)])
        return hist

    def update_weights(self, z):
        """
        Given the allocation vector z, resample the weights parameter

        Parameters
        ----------
        z array of shape (nb_samples), type = np.int
          the allocation variable
        """
        pop = self.pop(z)
        weights = pop + self.prior_weights
        self.weights = np.random.dirichlet(weights)

    def update_means(self, x, z):
        """
        Given the allocation vector z,
        and the corresponding data x,
        resample the mean

        Parameters
        ----------
        x: array of shape (nb_samples,self.dim)
          the data used in the estimation process
        z: array of shape (nb_samples), type = np.int
          the corresponding classification
        """
        pop = self.pop(z)
        self.shrinkage = self.prior_shrinkage + pop
        empmeans = np.zeros(np.shape(self.means))
        prior_shrinkage = np.reshape(self.prior_shrinkage, (self.k, 1))
        shrinkage = np.reshape(self.shrinkage, (self.k, 1))

        for k in range(self.k):
            empmeans[k] = np.sum(x[z == k], 0)

        means = empmeans + self.prior_means * prior_shrinkage
        means /= shrinkage
        for k in range(self.k):
            self.means[k] = generate_normals(\
                means[k], self.precisions[k] * self.shrinkage[k])

    def update_precisions(self, x, z):
        """
        Given the allocation vector z,
        and the corresponding data x,
        resample the precisions

        Parameters
        ----------
        x array of shape (nb_samples,self.dim)
          the data used in the estimation process
        z array of shape (nb_samples), type = np.int
          the corresponding classification
        """
        pop = self.pop(z)
        self.dof = self.prior_dof + pop + 1
        rpop = pop + (pop == 0)
        self._detp = np.zeros(self.k)

        for k in range(self.k):
            # empirical means
            empmeans = np.sum(x[z == k], 0) / rpop[k]
            dm = np.reshape(empmeans - self.prior_means[k], (1, self.dim))

            # scatter
            dx = np.reshape(x[z == k] - empmeans, (pop[k], self.dim))
            scatter = np.dot(dx.T, dx)

            # bias
            addcov = np.dot(dm.T, dm) * self.prior_shrinkage[k]

            # covariance = prior term + scatter + bias
            covariance = self._inv_prior_scale[k] + scatter + addcov

            #precision
            scale = inv(covariance)
            self.precisions[k] = generate_Wishart(self.dof[k], scale)
            self._detp[k] = detsh(self.precisions[k])

    def update(self, x, z):
        """
        update function (draw a sample of the GMM parameters)

        Parameters
        ----------
        x array of shape (nb_samples,self.dim)
          the data used in the estimation process
        z array of shape (nb_samples), type = np.int
          the corresponding classification
        """
        self.update_weights(z)
        self.update_precisions(x, z)
        self.update_means(x, z)

    def sample_indicator(self, like):
        """
        sample the indicator from the likelihood

        Parameters
        ----------
        like: array of shape (nb_samples,self.k)
           component-wise likelihood

        Returns
        -------
        z: array of shape(nb_samples): a draw of the membership variable
        """
        tiny = 1 + 1.e-15
        like = (like.T / like.sum(1)).T
        like /= tiny
        z = multinomial(like)
        return z

    def sample(self, x, niter=1, mem=0, verbose=0):
        """
        sample the indicator and parameters

        Parameters
        ----------
        x array of shape (nb_samples,self.dim)
          the data used in the estimation process
        niter=1 : the number of iterations to perform
        mem=0: if mem, the best values of the parameters are computed
        verbose=0: verbosity mode

        Returns
        -------
        best_weights: array of shape (self.k)
        best_means: array of shape (self.k, self.dim)
        best_precisions: array of shape (self.k, self.dim, self.dim)
        possibleZ: array of shape (nb_samples, niter)
                   the z that give the highest posterior
                   to the data is returned first
        """
        self.check_x(x)
        if mem:
            possibleZ = - np.ones((x.shape[0], niter)).astype(np.int)

        score = - np.inf
        bpz = - np.inf

        for i in range(niter):
            like = self.likelihood(x)
            sll = np.mean(np.log(np.sum(like, 1)))
            sll += np.log(self.probability_under_prior())
            if sll > score:
                score = sll
                best_weights = self.weights.copy()
                best_means = self.means.copy()
                best_precisions = self.precisions.copy()

            z = self.sample_indicator(like)
            if mem:
                possibleZ[:, i] = z
            puz = sll # to save time
            self.update(x, z)
            if puz > bpz:
                ibz = i
                bpz = puz

        if mem:
            aux = possibleZ[:, 0].copy()
            possibleZ[:, 0] = possibleZ[:, ibz].copy()
            possibleZ[:, ibz] = aux
            return best_weights, best_means, best_precisions, possibleZ

    def sample_and_average(self, x, niter=1, verbose=0):
        """
        sample the indicator and parameters
        the average values for weights,means, precisions are returned

        Parameters
        ----------
        x = array of shape (nb_samples,dim)
          the data from which bic is computed
        niter=1: number of iterations

        Returns
        -------
        weights: array of shape (self.k)
        means: array of shape (self.k,self.dim)
        precisions:  array of shape (self.k,self.dim,self.dim)
                     or (self.k, self.dim)
                     these are the average parameters across samplings

        Notes
        -----
        All this makes sense only if no label switching as occurred so this is
        wrong in general (asymptotically).

        fix: implement a permutation procedure for components identification
        """
        aprec = np.zeros(np.shape(self.precisions))
        aweights = np.zeros(np.shape(self.weights))
        ameans = np.zeros(np.shape(self.means))

        for i in range(niter):
            like = self.likelihood(x)
            z = self.sample_indicator(like)
            self.update(x, z)
            aprec += self.precisions
            aweights += self.weights
            ameans += self.means
        aprec /= niter
        ameans /= niter
        aweights /= niter
        return aweights, ameans, aprec

    def probability_under_prior(self):
        """
        Compute the probability of the current parameters of self
        given the priors
        """
        p0 = 1
        p0 = dirichlet_eval(self.weights, self.prior_weights)
        for k in range(self.k):
            mp = np.reshape(self.precisions[k] * self.prior_shrinkage[k],
                            (self.dim, self.dim))
            p0 *= normal_eval(self.prior_means[k], mp, self.means[k])
            p0 *= wishart_eval(self.prior_dof[k], self.prior_scale[k],
                               self.precisions[k], dV=self._dets[k],
                               dW=self._detp[k], piV=self._inv_prior_scale[k])
        return p0

    def conditional_posterior_proba(self, x, z, perm=None):
        """
        Compute the probability of the current parameters of self
        given x and z

        Parameters
        ----------
        x: array of shape (nb_samples, dim),
           the data from which bic is computed
        z: array of shape (nb_samples), type = np.int,
           the corresponding classification
        perm: array ok shape(nperm, self.k),typ=np.int, optional
              all permutation of z under which things will be recomputed
              By default, no permutation is performed
        """
        pop = self.pop(z)
        rpop = (pop + (pop == 0)).astype(np.float)
        dof = self.prior_dof + pop + 1
        shrinkage = self.prior_shrinkage + pop
        weights = pop + self.prior_weights

        # initialize the porsterior proba
        if perm == None:
            pp = dirichlet_eval(self.weights, weights)
        else:
            pp = np.array([dirichlet_eval(self.weights[pj], weights)
                           for pj in perm])

        for k in range(self.k):
            m1 = np.sum(x[z == k], 0)

            #0. Compute the empirical means
            empmeans = m1 / rpop[k]

            #1. the precisions
            dx = np.reshape(x[z == k] - empmeans, (pop[k], self.dim))
            dm = np.reshape(empmeans - self.prior_means[k], (1, self.dim))
            addcov = np.dot(dm.T, dm) * self.prior_shrinkage[k]

            covariance = self._inv_prior_scale[k] + np.dot(dx.T, dx) + addcov
            scale = inv(covariance)
            _dets = detsh(scale)

            #2. the means
            means = m1 + self.prior_means[k] * self.prior_shrinkage[k]
            means /= shrinkage[k]

            #4. update the posteriors
            if perm == None:
                pp *= wishart_eval(
                    dof[k], scale, self.precisions[k],
                    dV=_dets, dW=self._detp[k], piV=covariance)
            else:
                for j, pj in enumerate(perm):
                    pp[j] *= wishart_eval(
                        dof[k], scale, self.precisions[pj[k]], dV=_dets,
                        dW=self._detp[pj[k]], piV=covariance)

            mp = scale * shrinkage[k]
            _dP = _dets * shrinkage[k] ** self.dim
            if perm == None:
                pp *= normal_eval(means, mp, self.means[k], dP=_dP)
            else:
                for j, pj in enumerate(perm):
                    pp[j] *= normal_eval(
                        means, mp, self.means[pj[k]], dP=_dP)

        return pp

    def evidence(self, x, z, nperm=0, verbose=0):
        """
        See bayes_factor(self, x, z, nperm=0, verbose=0)
        """
        return self.bayes_factor(self, x, z, nperm, verbose)

    def bayes_factor(self, x, z, nperm=0, verbose=0):
        """
        Evaluate the Bayes Factor of the current model using Chib's method

        Parameters
        ----------
        x: array of shape (nb_samples,dim)
           the data from which bic is computed
        z: array of shape (nb_samples), type = np.int
           the corresponding classification
        nperm=0: int
            the number of permutations to sample
            to model the label switching issue
            in the computation of the Bayes Factor
            By default, exhaustive permutations are used
        verbose=0: verbosity mode

        Returns
        -------
        bf (float) the computed evidence (Bayes factor)

        Notes
        -----
        See: Marginal Likelihood from the Gibbs Output
        Journal article by Siddhartha Chib;
        Journal of the American Statistical Association, Vol. 90, 1995
        """
        niter = z.shape[1]
        p = []
        perm = generate_perm(self.k)
        if nperm > perm.shape[0]:
            nperm = perm.shape[0]
        for i in range(niter):
            if nperm == 0:
                temp = self.conditional_posterior_proba(x, z[:, i], perm)
                p.append(temp.mean())
            else:
                drand = np.argsort(np.random.rand(perm.shape[0]))[:nperm]
                temp = self.conditional_posterior_proba(x, z[:, i],
                                                        perm[drand])
                p.append(temp.mean())

        p = np.array(p)
        mp = np.mean(p)
        p0 = self.probability_under_prior()
        like = self.likelihood(x)
        bf = np.log(p0) + np.sum(np.log(np.sum(like, 1))) - np.log(mp)

        if verbose:
            print np.log(p0), np.sum(np.log(np.sum(like, 1))), np.log(mp)
        return bf


# ---------------------------------------------------------
# --- Variational Bayes inference -------------------------
# ---------------------------------------------------------


class VBGMM(BGMM):
    """
    Subclass of Bayesian GMMs (BGMM)
    that implements Variational Bayes estimation of the parameters
    """

    def __init__(self, k=1, dim=1, means=None, precisions=None,
                 weights=None, shrinkage=None, dof=None):
        BGMM.__init__(self, k, dim, means, precisions, weights, shrinkage, dof)
        self.scale = self.precisions.copy()

    def _Estep(self, x):
        """VB-E step

        Parameters
        ----------
        x array of shape (nb_samples,dim)
          the data used in the estimation process

        Returns
        -------
        like: array of shape(nb_samples,self.k),
              component-wise likelihood
        """
        n = x.shape[0]
        like = np.zeros((n, self.k))
        from scipy.special import psi

        spsi = psi(np.sum(self.weights))
        for k in range(self.k):
            # compute the data-independent factor first
            w0 = psi(self.weights[k]) - spsi
            w0 += 0.5 * np.log(detsh(self.scale[k]))
            w0 -= self.dim * 0.5 / self.shrinkage[k]
            w0 += 0.5 * np.log(2) * self.dim
            for i in range(self.dim):
                w0 += 0.5 * psi((self.dof[k] - i) / 2)
            m = np.reshape(self.means[k], (1, self.dim))
            b = self.dof[k] * self.scale[k]
            q = np.sum(np.dot(m - x, b) * (m - x), 1)
            w = w0 - q / 2
            w -= 0.5 * np.log(2 * np.pi) * self.dim
            like[:, k] = np.exp(w)

        if like.min() < 0:
            raise ValueError('Likelihood cannot be negative')
        return like

    def evidence(self, x, like=None, verbose=0):
        """computation of evidence bound aka free energy

        Parameters
        ----------
        x array of shape (nb_samples,dim)
          the data from which evidence is computed
        like=None: array of shape (nb_samples, self.k), optional
                   component-wise likelihood
                   If None, it is recomputed
        verbose=0: verbosity model

        Returns
        -------
        ev (float) the computed evidence
        """
        from scipy.special import psi
        from numpy.linalg import inv
        tiny = 1.e-15
        if like == None:
            like = self._Estep(x)
            like = (like.T / np.maximum(like.sum(1), tiny)).T

        pop = like.sum(0)[:self.k]
        pop = np.reshape(pop, (self.k, 1))
        spsi = psi(np.sum(self.weights))
        empmeans = np.dot(like.T[:self.k], x) / np.maximum(pop, tiny)

        F = 0
        # start with the average likelihood term
        for k in range(self.k):
            # compute the data-independent factor first
            Lav = psi(self.weights[k]) - spsi
            Lav -= np.sum(like[:, k] * np.log(np.maximum(like[:, k], tiny))) \
                   / pop[k]
            Lav -= 0.5 * self.dim * np.log(2 * np.pi)
            Lav += 0.5 * np.log(detsh(self.scale[k]))
            Lav += 0.5 * np.log(2) * self.dim
            for i in range(self.dim):
                Lav += 0.5 * psi((self.dof[k] - i) / 2)
            Lav -= self.dim * 0.5 / self.shrinkage[k]
            Lav *= pop[k]
            empcov = np.zeros((self.dim, self.dim))
            dx = x - empmeans[k]
            empcov = np.dot(dx.T, like[:, k:k + 1] * dx)
            Lav -= 0.5 * np.trace(np.dot(empcov, self.scale[k] * self.dof[k]))
            F += Lav

        #then the KL divergences
        prior_covariance = np.array(self._inv_prior_scale)
        covariance = np.array([inv(self.scale[k]) for k in range(self.k)])
        Dklw = 0
        Dklg = 0
        Dkld = dkl_dirichlet(self.weights, self.prior_weights)
        for k in range(self.k):
            Dklw += dkl_wishart(self.dof[k], covariance[k],
                                self.prior_dof[k], prior_covariance[k])
            nc = self.scale[k] * (self.dof[k] * self.shrinkage[k])
            nc0 = self.scale[k] * (self.dof[k] * self.prior_shrinkage[k])
            Dklg += dkl_gaussian(self.means[k], nc, self.prior_means[k], nc0)
        Dkl = Dkld + Dklg + Dklw
        if verbose:
            print 'Lav', F, 'Dkl', Dkld, Dklg, Dklw
        F -= Dkl
        return F

    def _Mstep(self, x, like):
        """VB-M step

        Parameters
        ----------
        x: array of shape(nb_samples, self.dim)
           the data from which the model is estimated
        like: array of shape(nb_samples, self.k)
           the likelihood of the data under each class
        """
        from numpy.linalg import inv
        tiny = 1.e-15
        pop = like.sum(0)

        # shrinkage, weights,dof
        self.weights = self.prior_weights + pop
        pop = pop[0:self.k]
        like = like[:, :self.k]
        self.shrinkage = self.prior_shrinkage + pop
        self.dof = self.prior_dof + pop

        #reshape
        pop = np.reshape(pop, (self.k, 1))
        prior_shrinkage = np.reshape(self.prior_shrinkage, (self.k, 1))
        shrinkage = np.reshape(self.shrinkage, (self.k, 1))

        # means
        means = np.dot(like.T, x) + self.prior_means * prior_shrinkage
        self.means = means / shrinkage

        #precisions
        empmeans = np.dot(like.T, x) / np.maximum(pop, tiny)
        empcov = np.zeros(np.shape(self.prior_scale))
        for k in range(self.k):
            dx = x - empmeans[k]
            empcov[k] = np.dot(dx.T, like[:, k:k + 1] * dx)
            covariance = np.array(self._inv_prior_scale) + empcov

        dx = np.reshape(empmeans - self.prior_means, (self.k, self.dim, 1))
        addcov = np.array([np.dot(dx[k], dx[k].T) for k in range(self.k)])
        apms = np.reshape(prior_shrinkage * pop / shrinkage, (self.k, 1, 1))
        covariance += addcov * apms
        # update scale
        self.scale = np.array([inv(covariance[k]) for k in range(self.k)])

    def initialize(self, x):
        """
        initialize z using a k-means algorithm, then upate the parameters

        Parameters
        ----------
        x: array of shape (nb_samples,self.dim)
           the data used in the estimation process
        """
        n = x.shape[0]
        if self.k > 1:
            cent, z, J = kmeans(x, self.k)
        else:
            z = np.zeros(x.shape[0]).astype(np.int)
        l = np.zeros((n, self.k))
        l[np.arange(n), z] = 1
        self._Mstep(x, l)

    def map_label(self, x, like=None):
        """
        return the MAP labelling of x

        Parameters
        ----------
        x array of shape (nb_samples,dim)
          the data under study
        like=None array of shape(nb_samples,self.k)
               component-wise likelihood
               if like==None, it is recomputed

        Returns
        -------
        z: array of shape(nb_samples): the resulting MAP labelling
           of the rows of x
        """
        if like == None:
            like = self.likelihood(x)
        z = np.argmax(like, 1)
        return z

    def estimate(self, x, niter=100, delta=1.e-4, verbose=0):
        """estimation of self given x

        Parameters
        ----------
        x array of shape (nb_samples,dim)
          the data from which the model is estimated
        z = None: array of shape (nb_samples)
          a prior labelling of the data to initialize the computation
        niter=100: maximal number of iterations in the estimation process
        delta = 1.e-4: increment of data likelihood at which
              convergence is declared
        verbose=0:
                verbosity mode
        """
        # alternation of E/M step until convergence
        tiny = 1.e-15
        av_ll_old = - np.inf
        for i in range(niter):
            like = self._Estep(x)
            av_ll = np.mean(np.log(np.maximum(np.sum(like, 1), tiny)))
            if av_ll < av_ll_old + delta:
                if verbose:
                    print 'iteration:', i, 'log-likelihood:', av_ll,\
                          'old value:', av_ll_old
                break
            else:
                av_ll_old = av_ll
            if verbose:
                print i, av_ll, self.bic(like)
            like = (like.T / np.maximum(like.sum(1), tiny)).T
            self._Mstep(x, like)

    def likelihood(self, x):
        """
        return the likelihood of the model for the data x
        the values are weighted by the components weights

        Parameters
        ----------
        x: array of shape (nb_samples, self.dim)
           the data used in the estimation process

        Returns
        -------
        like: array of shape(nb_samples, self.k)
              component-wise likelihood
        """
        x = self.check_x(x)
        return self._Estep(x)

    def pop(self, like, tiny=1.e-15):
        """
        compute the population, i.e. the statistics of allocation

        Parameters
        ----------
        like array of shape (nb_samples, self.k):
          the likelihood of each item being in each class
        """
        slike = np.maximum(tiny, np.sum(like, 1))
        nlike = (like.T / slike).T
        return np.sum(nlike, 0)



#!/usr/bin/env python

import numpy as np
from scipy.stats import norm

from ..utils import multiple_mahalanobis, z_score, multiple_fast_inv
from nose.tools import assert_true
from numpy.testing import assert_almost_equal, assert_array_almost_equal

def test_z_score():
    p = np.random.rand(10)
    z = z_score(p)
    assert_array_almost_equal(norm.sf(z), p) 

def test_mahalanobis():
    x = np.random.rand(100) / 100
    A = np.random.rand(100, 100) / 100
    A = np.dot(A.transpose(), A) + np.eye(100)
    mah = np.dot(x, np.dot(np.linalg.inv(A), x))
    assert_almost_equal(mah, multiple_mahalanobis(x, A), decimal=1) 

def test_mahalanobis2():
    x = np.random.randn(100, 3)
    Aa = np.zeros([100, 100, 3])
    for i in range(3):
        A = np.random.randn(120, 100)
        A = np.dot(A.T, A)
        Aa[:, :, i] = A
    i = np.random.randint(3)
    mah = np.dot(x[:, i], np.dot(np.linalg.inv(Aa[:, :, i]), x[:, i]))
    f_mah = (multiple_mahalanobis(x, Aa))[i]
    assert_true(np.allclose(mah, f_mah))

def test_multiple_fast_inv():
    shape = (10, 20, 20)
    X = np.random.randn(shape[0], shape[1], shape[2])
    X_inv_ref = np.zeros(shape)
    for i in range(shape[0]):
        X[i] = np.dot(X[i], X[i].T)
        X_inv_ref[i] = np.linalg.inv(X[i])
    X_inv = multiple_fast_inv(X)
    assert_almost_equal(X_inv_ref, X_inv)


if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])



#!/usr/bin/env python

"""
Functions to do automatic visualization of activation-like maps.

For 2D-only visualization, only matplotlib is required.
For 3D visualization, Mayavi, version 3.0 or greater, is required.

For a demo, see the 'demo_plot_map' function.

"""

# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# License: BSD

# Standard library imports
import warnings
import operator

# Standard scientific libraries imports (more specific imports are
# delayed, so that the part module can be used without them).
import numpy as np

# Import pylab
from nipy.utils.skip_test import skip_if_running_nose

try:
    import pylab as pl
except ImportError:
    skip_if_running_nose('Could not import matplotlib')

from .anat_cache import mni_sform, mni_sform_inv, _AnatCache
from .coord_tools import coord_transform

from .slicers import SLICERS, _xyz_order
from edge_detect import _fast_abs_percentile

################################################################################
# Helper functions for 2D plotting of activation maps
################################################################################


def plot_map(map, affine, cut_coords=None, anat=None, anat_affine=None,
                    slicer='ortho', figure=None, axes=None, title=None,
                    threshold=None, annotate=True, draw_cross=True,
                    do3d=False, threshold_3d=None,
                    view_3d=(38.5, 70.5, 300, (-2.7, -12, 9.1)),
                    black_bg=False, **kwargs):
    """ Plot three cuts of a given activation map (Frontal, Axial, and Lateral)

        Parameters
        ----------
        map : 3D ndarray
            The activation map, as a 3D image.
        affine : 4x4 ndarray
            The affine matrix going from image voxel space to MNI space.
        cut_coords: None, or a tuple of floats
            The MNI coordinates of the point where the cut is performed, in
            MNI coordinates and order.
            If slicer is 'ortho', this should be a 3-tuple: (x, y, z)
            For slicer == 'x', 'y', or 'z', then these are the
            coordinates of each cut in the corresponding direction.
            If None is given, the cuts is calculated automaticaly.
        anat : 3D ndarray or False, optional
            The anatomical image to be used as a background. If None, the
            MNI152 T1 1mm template is used. If False, no anat is displayed.
        anat_affine : 4x4 ndarray, optional
            The affine matrix going from the anatomical image voxel space to 
            MNI space. This parameter is not used when the default 
            anatomical is used, but it is compulsory when using an
            explicite anatomical image.
        slicer: {'ortho', 'x', 'y', 'z'}
            Choose the direction of the cuts. With 'ortho' three cuts are
            performed in orthogonal directions
        figure : integer or matplotlib figure, optional
            Matplotlib figure used or its number. If None is given, a
            new figure is created.
        axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
            The axes, or the coordinates, in matplotlib figure space,
            of the axes used to display the plot. If None, the complete
            figure is used.
        title : string, optional
            The title dispayed on the figure.
        threshold : a number, None, or 'auto'
            If None is given, the maps are not thresholded.
            If a number is given, it is used to threshold the maps:
            values below the threshold are plotted as transparent. If
            auto is given, the threshold is determined magically by
            analysis of the map.
        annotate: boolean, optional
            If annotate is True, positions and left/right annotation
            are added to the plot.
        draw_cross: boolean, optional
            If draw_cross is True, a cross is drawn on the plot to
            indicate the cut plosition.
        do3d: {True, False or 'interactive'}, optional
            If True, Mayavi is used to plot a 3D view of the
            map in addition to the slicing. If 'interactive', the
            3D visualization is displayed in an additional interactive
            window.
        threshold_3d:
            The threshold to use for the 3D view (if any). Defaults to
            the same threshold as that used for the 2D view.
        view_3d: tuple,
            The view used to take the screenshot: azimuth, elevation,
            distance and focalpoint, see the docstring of mlab.view.
        black_bg: boolean, optional
            If True, the background of the image is set to be black. If
            you whish to save figures with a black background, you
            will need to pass "facecolor='k', edgecolor='k'" to pylab's
            savefig.
        kwargs: extra keyword arguments, optional
            Extra keyword arguments passed to pylab.imshow

        Notes
        -----
        Arrays should be passed in numpy convention: (x, y, z)
        ordered.

        Use masked arrays to create transparency:

            import numpy as np
            map = np.ma.masked_less(map, 0.5)
            plot_map(map, affine)
    """

    map, affine = _xyz_order(map, affine)

    nan_mask = np.isnan(np.asarray(map))
    if np.any(nan_mask):
        map = map.copy()
        map[nan_mask] = 0
    del nan_mask

    # Deal with automatic settings of plot parameters
    if threshold == 'auto':
        # Threshold epsilon above a percentile value, to be sure that some
        # voxels are indeed threshold
        threshold = _fast_abs_percentile(map) + 1e-5

    if do3d:
        try:
            try:
                from mayavi import version
            except ImportError:
                from enthought.mayavi import version
            if not int(version.version[0]) > 2:
                raise ImportError
        except ImportError:
            warnings.warn('Mayavi > 3.x not installed, plotting only 2D')
            do3d = False

    slicer = SLICERS[slicer].init_with_figure(data=map, affine=affine,
                                          threshold=threshold,
                                          cut_coords=cut_coords,
                                          figure=figure, axes=axes,
                                          black_bg=black_bg,
                                          leave_space=do3d)

    # Use Mayavi for the 3D plotting
    if do3d:
        from .maps_3d import plot_map_3d, m2screenshot
        try:
            from tvtk.api import tvtk
        except ImportError:
            from enthought.tvtk.api import tvtk
        version = tvtk.Version()
        offscreen = True
        if (version.vtk_major_version, version.vtk_minor_version) < (5, 2):
            offscreen = False
        if do3d == 'interactive':
            offscreen = False

        cmap = kwargs.get('cmap', pl.cm.cmap_d[pl.rcParams['image.cmap']])
        # Computing vmin and vmax is costly in time, and is needed
        # later, so we compute them now, and store them for future
        # use
        vmin = kwargs.get('vmin', map.min())
        kwargs['vmin'] = vmin
        vmax = kwargs.get('vmax', map.max())
        kwargs['vmax'] = vmax
        try:
            from mayavi import mlab
        except ImportError:
            from enthought.mayavi import mlab
        if threshold_3d is None:
            threshold_3d = threshold
        plot_map_3d(np.asarray(map), affine, cut_coords=cut_coords,
                    anat=anat, anat_affine=anat_affine, 
                    offscreen=offscreen, cmap=cmap,
                    threshold=threshold_3d,
                    view=view_3d,
                    vmin=vmin, vmax=vmax)

        ax = slicer.axes.values()[0].ax.figure.add_axes((0.001, 0, 0.29, 1))
        ax.axis('off')
        m2screenshot(mpl_axes=ax)
        if offscreen:
            # Clean up, so that the offscreen engine doesn't become the
            # default
            mlab.clf()
            engine = mlab.get_engine()
            try:
                from mayavi.core.registry import registry
            except:
                from enthought.mayavi.core.registry import registry
            for key, value in registry.engines.iteritems():
                if value is engine:
                    registry.engines.pop(key)
                    break


    if threshold:
        map = np.ma.masked_inside(map, -threshold, threshold, copy=False)


    _plot_anat(slicer, anat, anat_affine, title=title,
               annotate=annotate, draw_cross=draw_cross)

    slicer.plot_map(map, affine, **kwargs)
    return slicer


def _plot_anat(slicer, anat, anat_affine, title=None,
               annotate=True, draw_cross=True, dim=False, cmap=pl.cm.gray):
    """ Internal function used to plot anatomy
    """
    canonical_anat = False
    if anat is None:
        try:
            anat, anat_affine, vmax_anat = _AnatCache.get_anat()
            canonical_anat = True
        except OSError, e:
            anat = False
            warnings.warn(repr(e))

    black_bg = slicer._black_bg
    # XXX: Check that we should indeed plot an anat: we have one, and the
    # cut_coords are in its range

    if anat is not False:
        if canonical_anat:
            # We special-case the 'canonical anat', as we don't need
            # to do a few transforms to it.
            vmin = 0
            vmax = vmax_anat
        elif dim:
            vmin = anat.min()
            vmax = anat.max()
        else:
            vmin = None
            vmax = None
            anat, anat_affine = _xyz_order(anat, anat_affine)
        if dim:
            vmean = .5*(vmin + vmax)
            ptp = .5*(vmax - vmin)
            if not operator.isNumberType(dim):
                dim = .6
            if black_bg:
                vmax = vmean + (1+dim)*ptp
            else:
                vmin = vmean - (1+dim)*ptp
        slicer.plot_map(anat, anat_affine, cmap=cmap,
                              vmin=vmin, vmax=vmax)

        if annotate:
            slicer.annotate()
        if draw_cross:
            slicer.draw_cross()

    if black_bg:
        # To have a black background in PDF, we need to create a
        # patch in black for the background
        for ax in slicer.axes.values():
            ax.ax.imshow(np.zeros((2, 2, 3)),
                         extent=[-5000, 5000, -5000, 5000],
                         zorder=-500)

    if title is not None and not title == '':
        slicer.title(title)
    return slicer


def plot_anat(anat=None, anat_affine=None, cut_coords=None, slicer='ortho',
              figure=None, axes=None, title=None, annotate=True,
              draw_cross=True, black_bg=False, dim=False, cmap=pl.cm.gray):
    """ Plot three cuts of an anatomical image (Frontal, Axial, and Lateral)

        Parameters
        ----------
        anat : 3D ndarray, optional
            The anatomical image to be used as a background. If None is
            given, nipy tries to find a T1 template.
        anat_affine : 4x4 ndarray, optional
            The affine matrix going from the anatomical image voxel space to 
            MNI space. This parameter is not used when the default 
            anatomical is used, but it is compulsory when using an
            explicite anatomical image.
        figure : integer or matplotlib figure, optional
            Matplotlib figure used or its number. If None is given, a
            new figure is created.
        cut_coords: None, or a tuple of floats
            The MNI coordinates of the point where the cut is performed, in
            MNI coordinates and order.
            If slicer is 'ortho', this should be a 3-tuple: (x, y, z)
            For slicer == 'x', 'y', or 'z', then these are the
            coordinates of each cut in the corresponding direction.
            If None is given, the cuts is calculated automaticaly.
        slicer: {'ortho', 'x', 'y', 'z'}
            Choose the direction of the cuts. With 'ortho' three cuts are
            performed in orthogonal directions
        figure : integer or matplotlib figure, optional
            Matplotlib figure used or its number. If None is given, a
            new figure is created.
        axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
            The axes, or the coordinates, in matplotlib figure space,
            of the axes used to display the plot. If None, the complete
            figure is used.
        title : string, optional
            The title dispayed on the figure.
        annotate: boolean, optional
            If annotate is True, positions and left/right annotation
            are added to the plot.
        draw_cross: boolean, optional
            If draw_cross is True, a cross is drawn on the plot to
            indicate the cut plosition.
        black_bg: boolean, optional
            If True, the background of the image is set to be black. If
            you whish to save figures with a black background, you
            will need to pass "facecolor='k', edgecolor='k'" to pylab's
            savefig.
        cmap: matplotlib colormap, optional
            The colormap for the anat

        Notes
        -----
        Arrays should be passed in numpy convention: (x, y, z)
        ordered.
    """
    slicer = SLICERS[slicer].init_with_figure(data=anat, affine=anat_affine,
                                          threshold=0, cut_coords=cut_coords,
                                          figure=figure, axes=axes,
                                          black_bg=black_bg)

    _plot_anat(slicer, anat, anat_affine, title=title,
               annotate=annotate, draw_cross=draw_cross, dim=dim, cmap=cmap)
    return slicer


def demo_plot_map(do3d=False, **kwargs):
    """ Demo activation map plotting.
    """
    map = np.zeros((182, 218, 182))
    # Color a asymetric rectangle around Broca area:
    x, y, z = -52, 10, 22
    x_map, y_map, z_map = coord_transform(x, y, z, mni_sform_inv)
    # Compare to values obtained using fslview. We need to add one as
    # voxels do not start at 0 in fslview.
    assert x_map == 142
    assert y_map +1 == 137
    assert z_map +1 == 95
    map[x_map-5:x_map+5, y_map-3:y_map+3, z_map-10:z_map+10] = 1
    return plot_map(map, mni_sform, threshold='auto',
                        title="Broca's area", do3d=do3d,
                        **kwargs)





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from .routines import (quantile, median, mahalanobis, svd, permutations,
                       combinations, gamln, psi)
from .zscore import zscore

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Computation of the canonical HRF used in fMRIstat, both the 2-term
spectral approximation and the Taylor series approximation, to a shifted
version of the canonical Glover HRF.

References
----------
Liao, C.H., Worsley, K.J., Poline, J-B., Aston, J.A.D., Duncan, G.H.,
    Evans, A.C. (2002). \'Estimating the delay of the response in fMRI
    data.\' NeuroImage, 16:593-606.
"""

import numpy as np
import numpy.linalg as npl
from scipy.interpolate import interp1d

from nipy.fixes.sympy.utilities.lambdify import implemented_function

from ..utils import T, lambdify_t
from .. import hrf
from .invert import invertR

def spectral_decomposition(hrf2decompose,
                           time=None,
                           delta=None,
                           ncomp=2):
    """ PCA decomposition of symbolic HRF shifted over time

    Perform a PCA expansion of a symbolic HRF, time shifted over the
    values in delta, returning the first ncomp components.

    This smooths out the HRF as compared to using a Taylor series
    approximation.

    Parameters
    ----------
    hrf2decompose : sympy expression
        An expression that can be lambdified as a function of 't'. This
        is the HRF to be expanded in PCA
    time : None or np.ndarray, optional
        None gives default value of np.linspace(-15,50,3251) chosen to
        match fMRIstat implementation.  This corresponds to a time
        interval of 0.02.  Presumed to be equally spaced.
    delta : None or np.ndarray, optional
        None results in default value of np.arange(-4.5, 4.6, 0.1)
        chosen to match fMRIstat implementation.
    ncomp : int, optional
        Number of principal components to retain.

    Returns
    -------
    hrf : [sympy expressions]
        A sequence length `ncomp` of symbolic HRFs that are the
        principal components.
    approx : 
        TODO
    """
    if time is None:
        time = np.linspace(-15,50,3251)
    dt = time[1] - time[0]
    if delta is None:
        delta = np.arange(-4.5, 4.6, 0.1)
    # make numerical implementation from hrf function and symbol t.
    # hrft returns function values when called with values for time as
    # input.
    hrft = lambdify_t(hrf2decompose(T))
    # Create stack of time-shifted HRFs.  Time varies over row, delta
    # over column.
    ts_hrf_vals = np.array([hrft(time - d) for d in delta]).T
    ts_hrf_vals = np.nan_to_num(ts_hrf_vals)
    # PCA 
    U, S, V = npl.svd(ts_hrf_vals, full_matrices=0)
    # make interpolators from the generated bases
    basis = []
    for i in range(ncomp):
        b = interp1d(time, U[:, i], bounds_error=False, fill_value=0.)
        # normalize components witn integral of abs of first component
        if i == 0: 
            d = np.fabs((b(time) * dt).sum())
        b.y /= d
        basis.append(b)
    # reconstruct time courses for all bases
    W = np.array([b(time) for b in basis]).T
    # regress basis time courses against original time shifted time
    # courses, ncomps by len(delta) parameter matrix
    WH = np.dot(npl.pinv(W), ts_hrf_vals)
    # put these into interpolators to get estimated coefficients for any
    # value of delta
    coef = [interp1d(delta, w, bounds_error=False, fill_value=0.) for w in WH]
    # swap sign of first component to match that of input HRF.  Swap
    # other components if we swap the first, to standardize signs of
    # components across SVD implementations.
    if coef[0](0) < 0: # coefficient at time shift of 0
        for i in range(ncomp):
            coef[i].y *= -1.
            basis[i].y *= -1.

    def approx(time, delta):
        value = 0
        for i in range(ncomp):
            value += coef[i](delta) * basis[i](time)
        return value

    approx.coef = coef
    approx.components = basis
    (approx.theta,
     approx.inverse,
     approx.dinverse,
     approx.forward,
     approx.dforward) = invertR(delta, approx.coef)
    # construct aliased functions from bases
    symbasis = []
    for i, b in enumerate(basis):
        symbasis.append(
            implemented_function('%s%d' % (str(hrf2decompose), i), b))
    return symbasis, approx


def taylor_approx(hrf2decompose,
                  time=None,
                  delta=None):
    """ A Taylor series approximation of an HRF shifted by times `delta`

    Returns original HRF and gradient of HRF

    Parameters
    ----------
    hrf2decompose : sympy expression
        An expression that can be lambdified as a function of 't'. This
        is the HRF to be expanded in PCA
    time : None or np.ndarray, optional
        None gives default value of np.linspace(-15,50,3251) chosen to
        match fMRIstat implementation.  This corresponds to a time
        interval of 0.02.  Presumed to be equally spaced.
    delta : None or np.ndarray, optional
        None results in default value of np.arange(-4.5, 4.6, 0.1)
        chosen to match fMRIstat implementation.

    Returns
    -------
    hrf : [sympy expressions]
        Sequence length 2 comprising (`hrf2decompose`, ``dhrf``) where
        ``dhrf`` is the first derivative of `hrf2decompose`.
    approx : 
        TODO

    References
    ----------
    Liao, C.H., Worsley, K.J., Poline, J-B., Aston, J.A.D., Duncan, G.H.,
    Evans, A.C. (2002). \'Estimating the delay of the response in fMRI
    data.\' NeuroImage, 16:593-606.
    """
    if time is None:
        time = np.linspace(-15,50,3251)
    dt = time[1] - time[0]
    if delta is None:
        delta = np.arange(-4.5, 4.6, 0.1)
    # make numerical implementation from hrf function and symbol t.
    # hrft returns function values when called with values for time as
    # input.
    hrft = lambdify_t(hrf2decompose(T))
    # interpolator for negative gradient of hrf
    dhrft = interp1d(time, -np.gradient(hrft(time), dt), bounds_error=False,
                    fill_value=0.)
    dhrft.y *= 2
    # Create stack of time-shifted HRFs.  Time varies over row, delta
    # over column.
    ts_hrf_vals = np.array([hrft(time - d) for d in delta]).T
    # hrf, dhrf
    W = np.array([hrft(time), dhrft(time)]).T
    # regress hrf, dhrf at times against stack of time-shifted hrfs
    WH = np.dot(npl.pinv(W), ts_hrf_vals)
    # put these into interpolators to get estimated coefficients for any
    # value of delta
    coef = [interp1d(delta, w, bounds_error=False,
                     fill_value=0.) for w in WH]
            
    def approx(time, delta):
        value = (coef[0](delta) * hrft(time)
                 + coef[1](delta) * dhrft(time))
        return value

    approx.coef = coef
    approx.components = [hrft, dhrft]
    (approx.theta,
     approx.inverse,
     approx.dinverse,
     approx.forward,
     approx.dforward) = invertR(delta, approx.coef)
    dhrf = implemented_function('d%s' % str(hrf2decompose), dhrft)
    return [hrf2decompose, dhrf], approx


canonical, canonical_approx = taylor_approx(hrf.glover)
spectral, spectral_approx = spectral_decomposition(hrf.glover)



import numpy as np
from ..forest import  Forest

def simple_forest():
    """ generate a simple forest
    """
    parents = np.array([2, 2, 4, 4, 4])
    F = Forest(5, parents)
    return F

def test_forest():
    """ test creation of forest object
    """
    F = simple_forest()
    assert F.E == 8
    assert F.cc().max() == 0

def test_forest_trivial():
    """ test creation of forest object
    """
    F = Forest(5)
    assert F.E == 0
    assert (F.cc() == np.arange(5)).all()

def test_children():
    """ test that we obtain children
    """
    sf = simple_forest()
    ch = sf.get_children()
    assert len(ch) == 5
    assert ch[0] == []
    assert ch[1] == []
    assert ch[2] == [0, 1]
    assert ch[3] == []
    assert ch[4] == [2, 3]
    
def test_descendants():
    """ test the get_descendants() method
    """
    sf = simple_forest()
    assert sf.get_descendants(0) == [0]
    assert sf.get_descendants(1) == [1]
    assert sf.get_descendants(2) == [0, 1, 2]
    assert sf.get_descendants(4) == [0, 1, 2, 3, 4]
    
def test_root():
    """ test the isroot() method
    """ 
    root = simple_forest().isroot()
    assert root[4] == True
    assert root.sum() == 1

def test_merge_simple_branches():
    """ test the merge_simple_branches() method
    """
    f = Forest(5, np.array([2, 2, 4, 4, 4])).merge_simple_branches()
    assert f.V == 5
    f = Forest(5, np.array([1, 2, 4, 4, 4])).merge_simple_branches()
    assert f.V == 3

def test_all_distances():
    """ test the all_distances() methods
    """
    f =  simple_forest()
    dg = f.all_distances()
    print dg
    assert dg[0, 3] == 3.
    assert dg.max() == 3.
    assert dg.min() == 0.
    assert dg.shape == (5, 5)
    dg = f.all_distances(1)
    assert dg[3] == 3.

def test_depth():
    """ test the depth_from_leaves() methods
    """
    f =  simple_forest()
    depth = f.depth_from_leaves()
    assert depth[0] == 0
    assert depth[1] == 0
    assert depth[3] == 0
    assert depth[2] == 1
    assert depth[4] == 2
    
def test_reorder():
    """ test the reorder_from_leaves_to_roots() method
    """
    f =  simple_forest()
    order = f.reorder_from_leaves_to_roots()
    assert (f.depth_from_leaves() == np.array([0, 0, 0, 1, 2])).all()
    assert (order == np.array([0, 1, 3, 2, 4])).all()

def test_leaves():
    """ test the leaves_of_a_subtree() method
    """
    f =  simple_forest()
    assert f.leaves_of_a_subtree([0, 1]) == True
    assert f.leaves_of_a_subtree([0, 3]) == False
    assert f.leaves_of_a_subtree([1, 3]) == False
    assert f.leaves_of_a_subtree([0, 1, 3]) == True
    assert f.leaves_of_a_subtree([1]) == True
   
def test_depth():
    """ Test the tree_depth() method
    """
    f =  simple_forest()
    assert f.tree_depth() == 3

def test_upward_and():
    """ test the propagate_upward_and() method
    """
    f =  simple_forest()
    assert(f.propagate_upward_and([0, 1, 0, 1, 0]) == [0, 1, 0, 1, 0]).all()
    assert(f.propagate_upward_and([0, 1, 1, 1, 0]) == [0, 1, 0, 1, 0]).all()
    assert(f.propagate_upward_and([0, 1, 1, 1, 1]) == [0, 1, 0, 1, 0]).all()
    assert(f.propagate_upward_and([1, 1, 0, 1, 0]) == [1, 1, 1, 1, 1]).all()

def test_upward():
    """ test the propagate_upward() method
    """
    f =  simple_forest()
    assert(f.propagate_upward([0, 0, 1, 3, 1]) == [0, 0, 0, 3, 1]).all()
    assert(f.propagate_upward([0, 0, 5, 0, 2]) == [0, 0, 0, 0, 0]).all()



if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
##############################################################################
# Random Thresholding Procedure (after M. Lavielle and C. Ludena)

import numpy as np
import scipy.stats as st
from nipy.algorithms.graph import wgraph_from_3d_grid
from ..group.routines import add_lines

tol = 1e-10

##############################################################################
# Wrappers


def randthresh_main(Y, K, XYZ=None, p=np.inf, varwind=False, knownull=True,
                    stop=False, verbose=False):
    """ Wrapper for random threshold functions

    Parameters
    ----------
    Y:  array of shape (n,),Observations
    K: int, Some positive integer
       (lower bound on the number of null hypotheses)
    XYZ: array of shape (3, n)  voxel coordinates.
         If not empty, connexity constraints are used on the non-null set
    p: float, optional, lp norm
    varwind: bool,
             Varying window variant (vs. fixed window, with width K)
    knownull: bool, optional,
              Known null distribution (observations assumed Exp(1) under H0)
              versus unknown (observations assumed Gaussian under H0)
    stop: bool, optional
          Stop when minimum is attained (save computation time)
    verbose: bool,  'Chatty' mode

    Returns
    -------
    D : dict
        containing the following fields:
        * "C"  (n-K) array  Lp norm of partial sums fluctuation
          about their conditional expectation
        * "thresh" <float> Detection threshold
        * "detect" (k,)    Index of detected activations

    Notes
    -----
    Random thresholding is performed only if null hypothesis of no activations
    is rejected at level 5%
    """
    if XYZ == None:
        return randthresh(Y, K, p, stop, verbose, varwind, knownull)
    else:
        return randthresh_connex(Y, K, XYZ, p, stop, verbose, varwind,
                                 knownull)


def randthresh(Y, K, p=np.inf, stop=False, verbose=False, varwind=False,
               knownull=True):
    """ Wrapper for random threshold functions (without connexity constraints)

    Parameters
    ----------
    Y: array of shape  (n,) Observations
    K: int,
       Some positive integer (lower bound on the number of null hypotheses)
    p: float, lp norm
    stop      <bool>  Stop when minimum is attained (save computation time)
    verbose   <bool>  'Chatty' mode
    varwind   <bool>  Varying window variant (vs. fixed window, with width K)
    knownull  <bool>
               Known null distribution (observations assumed Exp(1) under H0)
               versus unknown (observations assumed Gaussian under H0)
    Returns
    -------
    A dictionary D containing the following fields:
    "C" (n-K)
        Lp norm of partial sums fluctuation about their conditional expectation
    "thresh" <float> Detection threshold
    "detect" (k,)    Index of detected activations
    "v"      <float> Estimated null variance (if knownull is False)

    Notes
    -----
    Random thresholding is performed only if null hypothesis of no activations
    is rejected at level 5%
    """
    D = {}
    # Test presence of activity
    if knownull:
        X = Y
    else:
        v = np.square(Y).mean()
        X = np.clip( - np.log(1 - st.chi2.cdf(Y ** 2, 1, 0, scale=v)), 0,
                      1 / tol)
        D["v"] = v
    T = test_stat(X, p=np.inf)
    if T <= 0.65:
        print "No activity detected at 5% level"
        D["detect"] = np.array([])
        D["thresh"] = np.inf
    else:
        # Find optimal threshold
        if varwind:
            if knownull:
                C = randthresh_varwind_knownull(Y, K, p, stop, verbose)
            else:
                C, V = randthresh_varwind_gaussnull(
                    Y, K, p, stop, one_sided=False, verbose=verbose)
        else:
            if knownull:
                C = randthresh_fixwind_knownull(Y, K, p, stop, verbose)
            else:
                C, V = randthresh_fixwind_gaussnull(
                    Y, K, p, stop, one_sided=False, verbose=verbose)
        n = len(X)
        if stop:
            I = np.where(C > 0)[0]
            if len(I) > 0:
                ncoeffs = I[-1]
            else:
                ncoeffs = n - K
        else:
            I = np.where((C[2:] > C[1:-1]) * (C[1:-1] < C[:-2]))[0]
            if len(I) > 0:
                ncoeffs = I[np.argmin(C[1: -1][I])] + 1
            else:
                ncoeffs = n - K
        thresh = np.sort(np.abs(Y))[ - ncoeffs]
        # Detected activations
        detect = np.where(np.abs(Y) > thresh)[0]
        D["C"] = C[2:]
        D["thresh"] = thresh
        D["detect"] = detect
        if not knownull:
            D["v"] = V[2:]
    return D


def randthresh_connex(Y, K, XYZ, p=np.inf, stop=False, verbose=False,
                      varwind=False, knownull=True):
    """
    Wrapper for random threshold functions under connexity constraints

    Parameters
    ----------
    Y  (n,)    Observations
    K  <int>
       Some positive integer (lower bound on the number of null hypotheses)
    XYZ       (3,n)   voxel coordinates
    p         <float> lp norm
    stop      <bool>  Stop when minimum is attained (save computation time)
    verbose   <bool>  'Chatty' mode
    varwind   <bool>  Varying window variant (vs. fixed window, with width K)
    knownull  <bool>
              Known null distribution (observations assumed Exp(1) under H0)
              versus unknown (observations assumed Gaussian under H0)

    Returns
    -------
    A dictionary D containing the following fields:
    "C" (n-K)
        Lp norm of partial sums fluctuation about their conditional expectation
    "thresh"    <float>    Detection threshold
    "detect"    (ncoeffs,) Index of detected voxels

    Notes
    -----
    Random thresholding is performed only if null hypothesis of no activations
    is rejected at level 5%
    """
    # Test presence of activity
    D = {}
    if knownull:
        X = Y
    else:
        v = np.square(Y).mean()
        X = np.clip( - np.log(1 - st.chi2.cdf(Y ** 2, 1, 0, scale=v)),
                       0, 1 / tol)
        D["v"] = v
    T = test_stat(X, p=np.inf)
    if T <= 0.65:
        print "No activity detected at 5% level"
        D["detect"] = np.array([])
        D["thresh"] = np.inf
    else:
        # Find optimal threshold
        if varwind:
            if knownull:
                C = randthresh_varwind_knownull_connex(
                    Y, K, XYZ, p, stop, verbose)
            else:
                C, V = randthresh_varwind_gaussnull_connex(
                    Y, K, XYZ, p, stop, verbose=verbose)
        else:
            if knownull:
                C = randthresh_fixwind_knownull_connex(
                    Y, K, XYZ, p, stop, verbose)
            else:
                C, V = randthresh_fixwind_gaussnull_connex(
                    Y, K, XYZ, p, stop, verbose=verbose)
        n = len(X)
        if stop:
            I = np.where(C > 0)[0]
            if len(I) > 0:
                ncoeffs = I[-1]
            else:
                ncoeffs = n - K
        else:
            I = np.where((C[2:] > C[1:-1]) * (C[1:-1] < C[:-2]))[0]
            if len(I) > 0:
                ncoeffs = I[np.argmin(C[1:-1][I])] + 1
            else:
                ncoeffs = n - K
        thresh = np.sort(np.abs(Y))[ - ncoeffs]
        detect = np.where(np.abs(Y) > thresh)[0]

        # Remove isolated voxels
        iso = isolated(XYZ[:, detect])
        detect[iso] = -1
        detect = detect[detect != -1]
        D["C"] = C[2:]
        D["thresh"] = thresh
        D["detect"] = detect
        if knownull == False:
            Ynull = np.square(Y).copy()
            Ynull[detect] = np.nan
            Ynull = Ynull[np.isnan(Ynull) == False]
            D["v"] = V[2:]
    return D

#########################################################################
# random threshold functions without connexity constraints


def randthresh_fixwind_knownull(X, K, p=np.inf, stop=False, verbose=False):
    """Random threshold with fixed-window and known null distribution

    Parameters
    ==========
    X (n,): Observations (must be Exp(1) under H0)
    K <int>:
       Some positive integer (lower bound on the number of null hypotheses)
    p <float>: Lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K):
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    # Sort data
    sortX = np.sort(X)[:: - 1]
    C = np.zeros(n - K, float)
    T = np.cumsum(sortX)
    for k in xrange(2, n - K):
        #Ratio of expectations
        B = np.arange(1, K + 1) * (1 + I[:n - 1 - k].sum() - I[:K].cumsum())
        B /= float(K) * ( 1 + I[K: n - 1 - k].sum() )
        #Partial sums
        Tk = T[k + 1: k + K + 1] - T[k]
        #Conditional expectations
        Q = B * Tk[-1]
        if p == np.inf:
            C[k] = np.abs(Tk - Q).max() / np.sqrt(n)
        else:
            C[k] = ( np.abs(Tk - Q) ** p ).sum() / n ** (p / 2.0 + 1)
        if verbose:
            print "k :", k, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C


def randthresh_varwind_knownull(X, K, p=np.inf, stop=False, verbose=False):
    """Random threshold with varying window and known null distribution

    Parameters
    ==========
    X (n,): Observations (Exp(1) under H0)
    K <int>:
      Some positive integer (lower bound on the number of null hypotheses)
    p <float>: lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K)
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Sort data
    sortX = np.sort(X)[:: - 1]
    T = np.cumsum(sortX)
    C = np.zeros(n - K, float)
    for k in xrange(2, n - K):
        #Ratio of expectations
        B = np.arange(1, n - k) * ( 1 + I[:n - 1 - k].sum() -
                                    I[:n - k - 1].cumsum())
        B /= float(n - k - 1)
        #Partial sums
        Tk = T[k + 1:] - T[k]
        #Conditional expectations
        Q = B * Tk[ - 1]
        if p == np.inf:
            C[k] = np.abs(Tk - Q).max() / np.sqrt(n - k - 1)
        else:
            C[k] = ( np.abs(Tk - Q) ** p).sum() / (n - k - 1) ** (p / 2.0 + 1)
        if verbose:
            print "k:", k, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C


def randthresh_fixwind_gaussnull(Y, K, p=np.inf, stop=False, one_sided=False,
                                 verbose=False):
    """ Random threshold with fixed window and null gaussian distribution

    Parameters
    ==========
    Y array of shape (n,)
      Observations (assumed Gaussian under H0, with unknown variance)
    K, int,  Some positive integer
       (lower bound on the number of null hypotheses)
    p, float, lp norm
    stop: bool,
          Stop when minimum is attained (save computation time)
    one_sided: bool,
               If nonzero means are positive only (vs. positive or negative)

    Returns
    =======
    C array of shape (n-K)
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(Y)
    I = 1.0 / np.arange(1, n + 1)
    if one_sided:
        sortY = np.sort(Y)
        std = np.sqrt((np.sum(sortY[1:K] ** 2) + np.cumsum(sortY[K: n] ** 2))\
                          * 1.0 / np.arange(K, n))
        std = std[:: - 1]
    else:
        sortY = np.sort(np.square(Y))
        V = (np.sum(sortY[1: K]) + np.cumsum(sortY[K: n])) * \
            1.0 / np.arange(K, n)
        V = V[:: - 1]
    C = np.zeros(n - K, float)
    sortY = sortY[:: - 1]
    for k in xrange(2, n - K):
        if one_sided:
            X = np.clip( - np.log(1 - st.norm.cdf(sortY[k + 1: k + K + 1],
                                                  scale=std[k])), 0, 1 / tol)
        else:
            X = np.clip( -
                np.log(1 - st.chi2.cdf(sortY[k + 1: k + K + 1],
                                       1, 0, scale=V[k])), 0, 1 / tol)

        # Ratio of expectations
        B = np.arange(1, K + 1) * (1 + I[:n - 1 - k].sum() - I[: K].cumsum())
        B /= float(K) * (1 + I[K: n - 1 - k].sum())
        # Partial sums
        T = X.cumsum()
        # Conditional expectations
        Q = B * T[-1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n)
        else:
            C[k] = ( np.abs(T - Q) ** p ).sum() / n ** ( p / 2.0 + 1)
        if verbose:
            print "k:", k, "C[k]:", C[k]
        if C[k] > C[k-1] and C[k-1] < C[k-2] and stop:
            break
    return C, V


def randthresh_varwind_gaussnull(Y, K, p=np.inf, stop=False, one_sided=False,
                                 verbose=False):
    """Random threshold with fixed window and gaussian null distribution

    Parameters
    ==========
    Y  (n,) Observations (assumed Gaussian under H0, with unknown variance)
    K  <int>
       Some positive integer (lower bound on the number of null hypotheses)
    p         <float> lp norm
    stop      <bool>  Stop when minimum is attained (save computation time)
    one_sided <bool>
              If nonzero means are positive only (vs. positive or negative)

    Returns
    =======
    C (n-K)
       Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(Y)
    I = 1.0 / np.arange(1, n + 1)
    if one_sided:
        sortY = np.sort(Y)
        std = np.sqrt((np.sum(sortY[1: K] ** 2) + np.cumsum(sortY[K: n] ** 2))
                      * 1.0 / np.arange(K, n))
        std = std[:: - 1]
    else:
        sortY = np.sort(np.square(Y))
        V = (np.sum(sortY[1: K]) + np.cumsum(sortY[K: n]))\
            * 1.0 / np.arange(K, n)
        V = V[:: - 1]
    C = np.zeros(n - K, float)
    sortY = sortY[:: - 1]
    for k in xrange(2, n - K):
        if one_sided:
            X = np.clip( - np.log(1 - st.norm.cdf(sortY[k + 1:],
                                                  scale=std[k])), 0, 1 / tol)
        else:
            X = np.clip( -
                np.log(1 - st.chi2.cdf(sortY[k + 1:], 1, 0, scale=V[k])), 0,
                1 / tol)
        # Ratio of expectations
        B = np.arange(1, n - k) * ( 1 + I[: n - 1 - k].sum() - \
                                    I[: n - k - 1].cumsum() )
        B /= float(n - k - 1)
        # Partial sums
        T = X.cumsum()
        # Conditional expectations
        Q = B * T[ - 1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n)
        else:
            C[k] = ( np.abs(T - Q) ** p ).sum() / n ** (p / 2.0 + 1)
        if verbose:
            print "k:", k, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C, V

###############################################################################
# random threshold functions with connexity constraints


def randthresh_fixwind_knownull_connex(X, K, XYZ, p=np.inf, stop=False,
                                       verbose=False):
    """Random threshold with fixed-window and known null distribution,
    using connexity constraint on non-null set.

    Parameters
    ==========
    X (n,): Observations (must be Exp(1) under H0)
    XYZ (3,n): voxel coordinates
    K <int>:
      Some positive integer (lower bound on the number of null hypotheses)
    p <float>: Lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K):
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Sort data
    J = np.argsort(X)[:: - 1]
    sortX = X[J]
    C = np.zeros(n - K, float)
    T = np.zeros(K, float)
    L = np.zeros(n, int)
    L[J[0]] = 1
    for k in xrange(2, n - K):
        #Ratio of expectations
        B = np.arange(1, K + 1) * (1 + I[: n - 1 - k].sum() - I[: K].cumsum())
        B /= float(K) * ( 1 + I[K: n - 1 - k].sum() )
        Jk = J[:k]
        #Suprathreshold voxels connected to new voxel
        XYZk = np.abs(XYZ[:, Jk] - XYZ[:, J[k - 1]].reshape(3, 1))
        Lk = np.where((XYZk.sum(axis=0) <= 2) * (XYZk.max(axis=0) <= 1))[0]\
            [: - 1]
        if len(Lk) == 0:
            L[J[k - 1]] = 1
        else:
            L[J[Lk]] = 0
        Ik = np.where(L[Jk] == 1)[0]
        nk = len(Ik)
        #Partial sums
        if nk >= K:
            T = sortX[Ik[:K]].cumsum()
        elif nk == 0:
            T = sortX[k + 1: k + K + 1].cumsum()
        else:
            T[:nk] = sortX[Ik].cumsum()
            T[nk:] = T[nk - 1] + sortX[k + 1:k + K - nk + 1].cumsum()
        # Conditional expectations
        Q = B * T[-1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n)
        else:
            C[k] = (np.abs(T - Q) ** p).sum() / n ** (p / 2.0 + 1)
        if verbose:
            print "k:", k, "nk:", nk, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C


def randthresh_varwind_knownull_connex(X, K, XYZ, p=np.inf, stop=False,
                                       verbose=False):
    """Random threshold with varying window and known null distribution

    Parameters
    ==========
    X (n,): Observations (Exp(1) under H0)
    K <int>:
      Some positive integer (lower bound on the number of null hypotheses)
    XYZ (3,n): voxel coordinates
    p <float>: lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K)
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Sort data
    J = np.argsort(X)[:: - 1]
    sortX = X[J]
    C = np.zeros(n - K, float)
    L = np.zeros(n, int)
    L[J[0]] = 1
    for k in xrange(2, n - K):
        Jk = J[:k]
        #Suprathreshold voxels connected to new voxel
        XYZk = np.abs(XYZ[:, Jk] - XYZ[:, J[k-1]].reshape(3, 1))
        Lk = np.where((XYZk.sum(axis=0) <= 2) * (XYZk.max(axis=0) <= 1))\
            [0][:-1]
        if len(Lk) == 0:
            L[J[k - 1]] = 1
        else:
            L[J[Lk]] = 0
        Ik = np.where(L[Jk] == 1)[0]
        #Ik = isolated(XYZ[:, Jk])
        nk = len(Ik)
        #Ratio of expectations
        B = np.arange(1, n - k + nk) * ( 1 + I[:n - 1 - k + nk].sum() -
                                         I[:n - k - 1 + nk].cumsum())
        B /= float(n - k - 1 + nk)
        #Partial sums
        if nk == 0:
            T = sortX[k + 1:].cumsum()
        else:
            T = np.zeros(n - k + nk - 1, float)
            T[:nk] = sortX[Ik].cumsum()
            T[nk:] = T[nk - 1] + sortX[k + 1:].cumsum()

        #Conditional expectations
        Q = B * T[-1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n - k - 1 + nk)
        else:
            C[k] = ( np.abs(T - Q) ** p ).sum() / (n - k - 1 + nk) **\
                (p / 2.0 + 1)
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
        if verbose:
            print "k:", k, "nk:", nk, "C[k]:", C[k]
    return C


def randthresh_fixwind_gaussnull_connex(X, K, XYZ, p=np.inf, stop=False,
                                        verbose=False):
    """Random threshold with fixed-window and gaussian null distribution,
    using connexity constraint on non-null set.

    Parameters
    ==========
    X (n,): Observations (assumed Gaussian under H0)
    XYZ (3,n): voxel coordinates
    K <int>:
      Some positive integer (lower bound on the number of null hypotheses)
    p <float>: Lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K):
       Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Sort data
    J = np.argsort(X ** 2)[:: - 1]
    sortX = np.square(X)[J]
    C = np.zeros(n - K, float)
    V = np.zeros(n - K, float)
    T = np.zeros(K, float)
    L = np.zeros(n, int)
    L[J[0]] = 1
    for k in xrange(2, n - K):
        #Ratio of expectations
        B = np.arange(1, K + 1) * ( 1 + I[:n - 1 - k].sum() - I[:K].cumsum())
        B /= float(K) * ( 1 + I[K:n - 1 - k].sum())
        Jk = J[:k]
        #Suprathreshold voxels connected to new voxel
        XYZk = np.abs(XYZ[:, Jk] - XYZ[:, J[k - 1]].reshape(3, 1))
        Lk = np.where((XYZk.sum(axis=0) <= 2) *
                      (XYZk.max(axis=0) <= 1))[0][: - 1]
        if len(Lk) == 0:
            L[J[k - 1]] = 1
        else:
            L[J[Lk]] = 0
        Ik = np.where(L[Jk] == 1)[0]
        nk = len(Ik)
        #Null variance
        V[k] = (sortX[Ik].sum() + sortX[k + 1:].sum()) / float(nk + n - k - 1)
        #Partial sums
        if nk >= K:
            T = np.clip( - np.log(1 - st.chi2.cdf(
                        sortX[Ik[:K]], 1, 0, scale=V[k])), 0, 1 / tol).cumsum()
        elif nk == 0:
            T = np.clip( -
                np.log(1 - st.chi2.cdf(sortX[k + 1:k + K + 1], 1, 0,
                                       scale=V[k])), 0, 1 / tol).cumsum()
        else:
            T[:nk] = np.clip( -
                np.log(1 - st.chi2.cdf(sortX[Ik], 1, 0, scale=V[k])), 0,
                1 / tol).cumsum()
            T[nk:] = T[nk - 1] + np.clip( -
                np.log(1 - st.chi2.cdf(sortX[k + 1:k + K - nk + 1], 1, 0,
                                       scale=V[k])), 0, 1 / tol).cumsum()
        # Conditional expectations
        Q = B * T[ - 1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n)
        else:
            C[k] = (np.abs(T - Q) ** p).sum() / n ** (p / 2.0 + 1)
        if verbose:
            print "k:", k, "nk:", nk, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C, V


def randthresh_varwind_gaussnull_connex(X, K, XYZ, p=np.inf, stop=False,
                                        verbose=False):
    """Random threshold with fixed-window and gaussian null distribution,
    using connexity constraint on non-null set.

    Parameters
    ==========
    X (n,): Observations (assumed Gaussian under H0)
    XYZ (3,n): voxel coordinates
    K <int>:
      Some positive integer (lower bound on the number of null hypotheses)
    p <float>: Lp norm
    stop <bool>: Stop when minimum is attained (save computation time)

    Returns
    =======
    C (n-K):
      Lp norm of partial sums fluctuation about their conditional expectation
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Sort data
    J = np.argsort(X ** 2)[:: - 1]
    sortX = np.square(X)[J]
    C = np.zeros(n - K, float)
    V = np.zeros(n - K, float)
    T = np.zeros(K, float)
    L = np.zeros(n, int)
    L[J[0]] = 1
    for k in xrange(2, n - K):
        Jk = J[:k]
        #Suprathreshold voxels connected to new voxel
        XYZk = np.abs(XYZ[:, Jk] - XYZ[:, J[k - 1]].reshape(3, 1))
        Lk = np.where((XYZk.sum(axis=0) <= 2) *
                      (XYZk.max(axis=0) <= 1))[0][: - 1]
        if len(Lk) == 0:
            L[J[k - 1]] = 1
        else:
            L[J[Lk]] = 0
        Ik = np.where(L[Jk] == 1)[0]
        #Ik = isolated(XYZ[:, Jk])
        nk = len(Ik)
        #Ratio of expectations
        B = np.arange(1, n - k + nk) * ( 1 + I[:n - 1 - k + nk].sum() -
                                         I[:n - k - 1 + nk].cumsum())
        B /= float(n - k - 1 + nk)
        #Null variance
        V[k] = (sortX[Ik].sum() + sortX[k + 1:].sum()) / float(nk + n - k - 1)
        #Partial sums
        if nk == 0:
            T = np.clip( -
                np.log(1 - st.chi2.cdf(sortX[k + 1:], 1, 0, scale=V[k])), 0,
                1 / tol).cumsum()
        else:
            T = np.zeros(n - k + nk - 1, float)
            T[:nk] = np.clip( -
                np.log(1 - st.chi2.cdf(sortX[Ik], 1, 0, scale=V[k])), 0,
                1 / tol).cumsum()
            T[nk:] = T[nk-1] + np.clip( -
                np.log(1 - st.chi2.cdf(sortX[k + 1:], 1, 0, scale=V[k])), 0,
                1 / tol).cumsum()
        #Conditional expectations
        Q = B * T[-1]
        if p == np.inf:
            C[k] = np.abs(T - Q).max() / np.sqrt(n - k - 1 + nk)
        else:
            C[k] = ( np.abs(T - Q) ** p).sum() / \
                (n - k - 1 + nk) ** (p / 2.0 + 1)
        if verbose:
            print "k:", k, "nk:", nk, "C[k]:", C[k]
        if C[k] > C[k - 1] and C[k - 1] < C[k - 2] and stop:
            break
    return C, V

#############################################################################
# Miscellanous functions


def test_stat(X, p=np.inf):
    """Test statistic of global null hypothesis
    that all observations have zero-mean

    Parameters
    ==========
    X (n,)    : X[j] = -log(1-F(|Y[j]|))
                where F: cdf of |Y[j]| under null hypothesis
                (must be computed beforehand)
    p         : Lp norm (<= inf) to use for computing test statistic

    Returns
    =======
    D <float> : test statistic
    """
    n = len(X)
    I = 1.0 / np.arange(1, n + 1)
    #Partial sums
    T = np.cumsum(np.sort(X)[:: - 1])
    #Expectation of partial sums
    E = np.arange(1, n + 1) * (1 + I.sum() - I.cumsum())
    #Conditional expectation of partial sums
    Q = E / n * T[ - 1]
    #Test statistic
    if p == np.inf:
        return np.max( ( np.abs(T - Q) ) / np.sqrt(n) )
    else:
        return sum(np.abs(T - Q) ** p) / (n ** (0.5 * p + 1))


def isolated(XYZ, k=18):
    """
    Outputs an index I of isolated points from their integer coordinates,
    XYZ (3, n), and under k-connectivity, k = 6, 18 or 26.
    """
    label = wgraph_from_3d_grid(XYZ.T, k).cc()
    # Isolated points
    ncc = label.max() + 1
    p = XYZ.shape[1]
    size = np.zeros(ncc, float)
    ones = np.ones((p, 1), float)
    add_lines(ones, size.reshape(ncc, 1), label)
    return np.where(size[label] == 1)[0]



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
Example of a script that uses the BSA (Bayesian Structural Analysis) i.e.
nipy.labs.spatial_models.bayesian_structural_analysis module.

Author : Bertrand Thirion, 2008-2010
"""
print(__doc__)

#autoindent
from os import mkdir, getcwd, path
import pickle

from numpy import array
from scipy import stats

from nipy.labs.spatial_models.bsa_io import make_bsa_image

# Local import
from get_data_light import DATA_DIR, get_second_level_dataset

# Get the data
nbsubj = 12
nbeta = 29
data_dir = path.join(DATA_DIR, 'group_t_images')
mask_images = [path.join(data_dir, 'mask_subj%02d.nii' % n)
               for n in range(nbsubj)]

betas = [path.join(data_dir, 'spmT_%04d_subj_%02d.nii' % (nbeta, n))
         for n in range(nbsubj)]

missing_file = array([not path.exists(m) for m in mask_images + betas]).any()

if missing_file:
    get_second_level_dataset()

# set various parameters
subj_id = ['%04d' % i for i in range(12)]
theta = float(stats.t.isf(0.01, 100))
dmax = 4.
ths = 0
thq = 0.95
verbose = 1
smin = 5
write_dir = path.join(getcwd(), 'results')
if not path.exists(write_dir):
    mkdir(write_dir)
method = 'quick'
print('method used:', method)

# call the function
AF, BF = make_bsa_image(mask_images, betas, theta, dmax, ths, thq, smin,
                        write_dir, method, subj_id, '%04d' % nbeta,
                        reshuffle=False)

# Write the result. OK, this is only a temporary solution
picname = path.join(write_dir, "AF_%04d.pic" % nbeta)
pickle.dump(AF, open(picname, 'wb'), 2)
picname = path.join(write_dir, "BF_%04d.pic" % nbeta)
pickle.dump(BF, open(picname, 'wb'), 2)

print("Wrote all the results in directory %s" % write_dir)



# -*- Mode: Python -*-  Not really, but the syntax is close enough


"""
Iterators for testing. 
Author: Alexis Roche, 2009.
"""

__version__ = '0.1'


# Includes
from fff cimport *
from numpy cimport dtype
cimport numpy as cnp

# Initialize numpy
fffpy_import_array()
import_array()
import numpy as np


c_types = ['unknown type', 'unsigned char', 'signed char', 'unsigned short', 'signed short', 
           'int', 'unsigned int', 'unsigned long', 'long', 'float', 'double']

fff_types = [FFF_UNKNOWN_TYPE, FFF_UCHAR, FFF_SCHAR, FFF_USHORT, FFF_SSHORT, 
             FFF_UINT, FFF_INT, FFF_ULONG, FFF_LONG, FFF_FLOAT, FFF_DOUBLE]  

npy_types = [cnp.NPY_NOTYPE, cnp.NPY_UBYTE, cnp.NPY_BYTE, cnp.NPY_USHORT,
             cnp.NPY_SHORT, cnp.NPY_UINT, cnp.NPY_INT, cnp.NPY_ULONG,
             cnp.NPY_LONG, cnp.NPY_FLOAT, cnp.NPY_DOUBLE]


def fff_type(dtype T):
    """
    fff_t, nbytes = fff_type(T)

    T is a np.dtype instance. Return a tuple (str, int). 
    """
    cdef fff_datatype fff_t
    cdef unsigned int nbytes
    fff_t =  fff_datatype_fromNumPy(T.type_num)
    nbytes =  fff_nbytes(fff_t)
    return c_types[fff_types.index(<int>fff_t)], nbytes


def npy_type(T): 
    """
    npy_t, nbytes = npy_type(T)

    T is a string. Return a tuple (str, int). 
    """
    cdef int npy_t
    cdef fff_datatype fff_t
    cdef unsigned int nbytes
    fff_t = <fff_datatype>fff_types[c_types.index(T)]
    npy_t = fff_datatype_toNumPy(fff_t)
    nbytes =  fff_nbytes(fff_t)
    return c_types[npy_types.index(npy_t)], nbytes

def pass_vector(ndarray X):
    """
    Y = pass_vector(X)
    """
    cdef fff_vector *x, *y
    x = fff_vector_fromPyArray(X)
    y = fff_vector_new(x.size)
    fff_vector_memcpy(y, x)
    fff_vector_delete(x)
    return fff_vector_toPyArray(y)


def copy_vector(ndarray X, int flag): 
    """
    Y = copy_vector(X, flag)

    flag == 0 ==> use fff
    flag == 1 ==> use numpy
    """
    cdef fff_vector *y
    cdef void* data
    cdef int size, stride, relstride, type, itemsize
    cdef fff_datatype fff_type

    data = <void*>X.data 
    size = X.shape[0]
    stride = X.strides[0]
    itemsize = X.descr.elsize
    type = X.descr.type_num 
    
    relstride = stride/itemsize
    fff_type =  fff_datatype_fromNumPy(type)

    y = fff_vector_new(size)

    if flag == 0: 
        fff_vector_fetch(y, data, fff_type, relstride) 
    else: 
        fff_vector_fetch_using_NumPy(y, <char*>data, stride, type, itemsize)

    return fff_vector_toPyArray(y)


def pass_matrix(ndarray X):
    """
    Y = pass_matrix(X)
    """
    cdef fff_matrix *x, *y 
    x = fff_matrix_fromPyArray(X)
    y = fff_matrix_new(x.size1, x.size2)
    fff_matrix_memcpy(y, x)
    fff_matrix_delete(x)
    return fff_matrix_toPyArray(y)


def pass_array(ndarray X):
    """
    Y = pass_array(X)
    """
    cdef fff_array *x, *y
    x = fff_array_fromPyArray(X)
    y = fff_array_new(x.datatype, x.dimX, x.dimY, x.dimZ, x.dimT)
    fff_array_copy(y, x)
    fff_array_delete(x)
    return fff_array_toPyArray(y)


def pass_vector_via_iterator(ndarray X, int axis=0, int niters=0):
    """
    Y = pass_vector_via_iterator(X, axis=0, niters=0)
    """
    cdef fff_vector *x, *y, *z
    cdef fffpy_multi_iterator* multi

    Xdum = X.copy() ## at least two arrays needed for multi iterator
    multi = fffpy_multi_iterator_new(2, axis, <void*>X, <void*>Xdum)
    x = multi.vector[0]

    while(multi.index < niters): 
        fffpy_multi_iterator_update(multi)

    y = fff_vector_new(x.size)
    fff_vector_memcpy(y, x)
    fffpy_multi_iterator_delete(multi)
    return fff_vector_toPyArray(y)


def copy_via_iterators(ndarray Y, int axis=0): 
    """
    Z = copy_via_iterators(Y, int axis=0) 

    Copy array Y into Z via fff's PyArray_MultiIterAllButAxis C function.
    Behavior should be equivalent to Z = Y.copy(). 
    """
    cdef fff_vector *y, *z
    cdef fffpy_multi_iterator* multi
   
    # Allocate output array
    Z = np.zeros_like(Y, dtype=np.float)

    # Create a new array iterator 
    multi = fffpy_multi_iterator_new(2, axis, <void*>Y, <void*>Z)
    
    # Create views
    y = multi.vector[0]
    z = multi.vector[1]

    # Loop 
    while(multi.index < multi.size):
        fff_vector_memcpy(z, y)
        fffpy_multi_iterator_update(multi)
    
    # Free memory
    fffpy_multi_iterator_delete(multi)

    # Return
    return Z


def sum_via_iterators(ndarray Y, int axis=0): 
    """
    Z = dummy_iterator(Y, int axis=0) 

    Return the sum of input elements along the dimension specified by axis.
    Behavior should be equivalent to Z = Y.sum(axis). 
    """
    cdef fff_vector *y, *z
    cdef fffpy_multi_iterator* multi
   
    # Allocate output array
    dims = [Y.shape[i] for i in range(Y.ndim)]
    dims[axis] = 1
    Z = np.zeros(dims)

    # Create a new array iterator 
    multi = fffpy_multi_iterator_new(2, axis, <void*>Y, <void*>Z)

    # Create views
    y = multi.vector[0]
    z = multi.vector[1]

    # Loop 
    while(multi.index < multi.size):
        z.data[0] = <double>fff_vector_sum(y)
        fffpy_multi_iterator_update(multi)
    
    # Free memory
    fffpy_multi_iterator_delete(multi)

    # Return
    return Z.squeeze()





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The testing directory contains a small set of imaging files to be
used for doctests only.  More thorough tests and example data will be
stored in a nipy data packages that you can download separately.

.. note:

   We use the ``nose`` testing framework for tests.

   Nose is a dependency for the tests, but should not be a dependency
   for running the algorithms in the NIPY library.  This file should
   import without nose being present on the python path.

Examples
--------

>>> from nipy.testing import funcfile
>>> from nipy.io.api import load_image
>>> img = load_image(funcfile)
>>> img.shape
(17, 21, 3, 20)

"""

import os

#__all__ = ['funcfile', 'anatfile']

# Discover directory path
filepath = os.path.abspath(__file__)
basedir = os.path.dirname(filepath)

funcfile = os.path.join(basedir, 'functional.nii.gz')
anatfile = os.path.join(basedir, 'anatomical.nii.gz')

from numpy.testing import *
# Overwrites numpy.testing.Tester
from .nosetester import NipyNoseTester as Tester
test = Tester().test
bench = Tester().bench

from . import decorators as dec

# Allow failed import of nose if not now running tests
try:
    from nose.tools import assert_true, assert_false
except ImportError:
    pass



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
"""
Example of a script that perfoms histogram analysis of an activation image.
This is based on a real fMRI image.

Simply modify the input image path to make it work on your preferred image.

Needs matplotlib

Author : Bertrand Thirion, 2008-2009
"""

import os

import numpy as np

import scipy.stats as st

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nibabel import load

import nipy.algorithms.statistics.empirical_pvalue as en

# Local import
from get_data_light import DATA_DIR, get_second_level_dataset


# parameters
verbose = 1
theta = float(st.t.isf(0.01, 100))

# paths
mask_image = os.path.join(DATA_DIR, 'mask.nii.gz')
input_image = os.path.join(DATA_DIR, 'spmT_0029.nii.gz')
if (not os.path.exists(mask_image)) or (not os.path.exists(input_image)):
    get_second_level_dataset()

# Read the mask
nim = load(mask_image)
mask = nim.get_data()

# read the functional image
rbeta = load(input_image)
beta = rbeta.get_data()
beta = beta[mask > 0]

mf = plt.figure(figsize=(13, 5))
a1 = plt.subplot(1, 3, 1)
a2 = plt.subplot(1, 3, 2)
a3 = plt.subplot(1, 3, 3)

# fit beta's histogram with a Gamma-Gaussian mixture
bfm = np.array([2.5, 3.0, 3.5, 4.0, 4.5])
bfp = en.gamma_gaussian_fit(beta, bfm, verbose=1, mpaxes=a1)

# fit beta's histogram with a mixture of Gaussians
alpha = 0.01
pstrength = 100
bfq = en.three_classes_GMM_fit(beta, bfm, alpha, pstrength,
                               verbose=1, mpaxes=a2)

# fit the null mode of beta with the robust method
efdr = en.NormalEmpiricalNull(beta)
efdr.learn()
efdr.plot(bar=0, mpaxes=a3)

a1.set_title('Fit of the density with \n a Gamma-Gaussian mixture')
a2.set_title('Fit of the density with \n a mixture of Gaussians')
a3.set_title('Robust fit of the density \n with a single Gaussian')
plt.show()



#!/usr/bin/env python
import numpy


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('graph', parent_package, top_path)
    config.add_subpackage('tests')
    config.add_extension('_graph',
                         sources=['_graph.c'],
                         include_dirs=[numpy.get_include()])
    return config


if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Script to run the main analyses in parallel, using the IPython machinery.

See ``fiac_example.py``.
"""
#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import os

import numpy as np

from IPython import parallel

#-----------------------------------------------------------------------------
# Utility functions
#-----------------------------------------------------------------------------

_client = None
def setup_client():
    """Get a Client and initialize it.

    This assumes that all nodes see a shared filesystem.
    """
    global _client
    if _client is None:
        _client = parallel.Client()
        mydir = os.path.split(os.path.abspath(__file__))[0]
        def cd(path):
            import os
            os.chdir(path)
        _client[:].apply_sync(cd, mydir)
    return _client


def getruns():
    for i in range(16):
        for j in range(1,5):
            yield i, j

def getvals():
    for con in ['sentence:speaker_0',
                'sentence_1',
                'sentence_0',
                'sentence:speaker_1',
                'speaker_1',
                'speaker_0',
                'constant_1',
                'constant_0']:
        for design in ['block', 'event']:
            yield design, con

#-----------------------------------------------------------------------------
# Main analysis functions
#-----------------------------------------------------------------------------

def fitruns():
    """Run the basic model fit."""
    rc = setup_client()
    view = rc.load_balanced_view()
    i_s, j_s = zip(*getruns())

    def _fit(subj, run):
        import fiac_example
        try:
            return fiac_example.run_model(subj, run)
        except IOError:
            pass

    return view.map(_fit, i_s, j_s)


def fitfixed():
    """Run the fixed effects analysis for all subjects."""
    rc = setup_client()
    view = rc.load_balanced_view()
    subjects = range(16)

    def _fit(subject):
        import fiac_example
        try:
            fiac_example.fixed_effects(subject, "block")
        except IOError:
            pass
        try:
            fiac_example.fixed_effects(subject, "event")
        except IOError:
            pass

    return view.map(_fit, subjects)


def fitgroup():
    """Run the group analysis"""
    rc = setup_client()
    view = rc.load_balanced_view()
    d_s, c_s = zip(*getvals())

    def _fit(d, c):
        import fiac_example
        return fiac_example.group_analysis(d, c)

    return view.map(_fit, d_s, c_s)


def run_permute_test(design, contrast, nsample=1000):
    rc = setup_client()
    dview = rc[:]
    nnod = len(dview)
    # Samples per node.  Round up
    ns_nod = np.ceil(nsample / float(nnod))

    def _run_test(n, des, con):
        import fiac_example
        from fiac_example import GROUP_MASK
        min_vals, max_vals = fiac_example.permutation_test(des, con,
                                       GROUP_MASK, n)
        return min_vals, max_vals

    ar = dview.apply_async(_run_test, ns_nod, design, contrast)
    min_vals, max_vals = zip(*[r for r in ar])
    return np.concatenate(min_vals), np.concatenate(max_vals)


#-----------------------------------------------------------------------------
# Script entry point
#-----------------------------------------------------------------------------
if __name__ == '__main__':
    pass



'''Functions to operate on, or return, quaternions.

The module also includes functions for the closely related angle, axis
pair as a specification for rotation.

Quaternions here consist of 4 values ``w, x, y, z``, where ``w`` is the
real (scalar) part, and ``x, y, z`` are the complex (vector) part. 

Note - rotation matrices here apply to column vectors, that is,
they are applied on the left of the vector.  For example:

>>> import numpy as np
>>> q = [0, 1, 0, 0] # 180 degree rotation around axis 0
>>> M = quat2mat(q) # from this module
>>> vec = np.array([1, 2, 3]).reshape((3,1)) # column vector
>>> tvec = np.dot(M, vec)
'''

import math
import numpy as np

_MAX_FLOAT = np.maximum_sctype(np.float)
_FLOAT_EPS = np.finfo(np.float).eps


def fillpositive(xyz, w2_thresh=None):
    ''' Compute unit quaternion from last 3 values
   
    Parameters
    ----------
    xyz : iterable
       iterable containing 3 values, corresponding to quaternion x, y, z
    w2_thresh : None or float, optional
       threshold to determine if w squared is really negative.
       If None (default) then w2_thresh set equal to
       ``-np.finfo(xyz.dtype).eps``, if possible, otherwise
       ``-np.finfo(np.float).eps``

    Returns
    -------
    wxyz : array shape (4,)
         Full 4 values of quaternion

    Notes
    -----
    If w, x, y, z are the values in the full quaternion, assumes w is
    positive.

    Gives error if w*w is estimated to be negative

    w = 0 corresponds to a 180 degree rotation

    The unit quaternion specifies that np.dot(wxyz, wxyz) == 1.
    
    If w is positive (assumed here), w is given by:

    w = np.sqrt(1.0-(x*x+y*y+z*z))

    w2 = 1.0-(x*x+y*y+z*z) can be near zero, which will lead to
    numerical instability in sqrt.  Here we use the system maximum
    float type to reduce numerical instability

    Examples
    --------
    >>> import numpy as np
    >>> wxyz = fillpositive([0,0,0])
    >>> np.all(wxyz == [1, 0, 0, 0])
    True
    >>> wxyz = fillpositive([1,0,0]) # Corner case; w is 0
    >>> np.all(wxyz == [0, 1, 0, 0])
    True
    >>> np.dot(wxyz, wxyz)
    1.0
    '''
    # Check inputs (force error if < 3 values)
    if len(xyz) != 3:
        raise ValueError('xyz should have length 3')
    # If necessary, guess precision of input
    if w2_thresh is None:
        try: # trap errors for non-array, integer array
            w2_thresh = -np.finfo(xyz.dtype).eps
        except (AttributeError, ValueError):
            w2_thresh = -_FLOAT_EPS
    # Use maximum precision
    xyz = np.asarray(xyz, dtype=_MAX_FLOAT)
    # Calculate w
    w2 = 1.0 - np.dot(xyz, xyz)
    if w2 < 0:
        if w2 < w2_thresh:
            raise ValueError('w2 should be positive, but is %f' % w2)
        w = 0
    else:
        w = np.sqrt(w2)
    return np.r_[w, xyz]


def quat2mat(q):
    ''' Calculate rotation matrix corresponding to quaternion

    Parameters
    ----------
    q : 4 element array-like

    Returns
    -------
    M : (3,3) array
      Rotation matrix corresponding to input quaternion *q*

    Notes
    -----
    Rotation matrix applies to column vectors, and is applied to the
    left of coordinate vectors.  The algorithm here allows non-unit
    quaternions.

    References
    ----------
    Algorithm from 
    http://en.wikipedia.org/wiki/Rotation_matrix#Quaternion

    Examples
    --------
    >>> import numpy as np
    >>> M = quat2mat([1, 0, 0, 0]) # Identity quaternion
    >>> np.allclose(M, np.eye(3))
    True
    >>> M = quat2mat([0, 1, 0, 0]) # 180 degree rotn around axis 0
    >>> np.allclose(M, np.diag([1, -1, -1]))
    True
    '''
    w, x, y, z = q
    Nq = w*w + x*x + y*y + z*z
    if Nq == 0.0:
        return np.eye(3)
    s = 2.0/Nq
    X = x*s
    Y = y*s
    Z = z*s
    wX = w*X; wY = w*Y; wZ = w*Z
    xX = x*X; xY = x*Y; xZ = x*Z
    yY = y*Y; yZ = y*Z; zZ = z*Z
    return np.array(
           [[ 1.0-(yY+zZ), xY-wZ, xZ+wY ],
            [ xY+wZ, 1.0-(xX+zZ), yZ-wX ],
            [ xZ-wY, yZ+wX, 1.0-(xX+yY) ]])


def mat2quat(M):
    ''' Calculate quaternion corresponding to given rotation matrix

    Parameters
    ----------
    M : array-like
      3x3 rotation matrix

    Returns
    -------
    q : (4,) array
      closest quaternion to input matrix, having positive q[0]

    Notes
    -----    
    Method claimed to be robust to numerical errors in M

    Constructs quaternion by calculating maximum eigenvector for matrix
    K (constructed from input `M`).  Although this is not tested, a
    maximum eigenvalue of 1 corresponds to a valid rotation.

    A quaternion q*-1 corresponds to the same rotation as q; thus the
    sign of the reconstructed quaternion is arbitrary, and we return
    quaternions with positive w (q[0]).
    
    Examples
    --------
    >>> import numpy as np
    >>> q = mat2quat(np.eye(3)) # Identity rotation
    >>> np.allclose(q, [1, 0, 0, 0])
    True
    >>> q = mat2quat(np.diag([1, -1, -1]))
    >>> np.allclose(q, [0, 1, 0, 0]) # 180 degree rotn around axis 0
    True

    Notes
    -----
    http://en.wikipedia.org/wiki/Rotation_matrix#Quaternion

    Bar-Itzhack, Itzhack Y. (2000), "New method for extracting the
    quaternion from a rotation matrix", AIAA Journal of Guidance,
    Control and Dynamics 23(6):1085-1087 (Engineering Note), ISSN
    0731-5090

    '''
    # Qyx refers to the contribution of the y input vector component to
    # the x output vector component.  Qyx is therefore the same as
    # M[0,1].  The notation is from the Wikipedia article.
    Qxx,Qyx,Qzx,Qxy,Qyy,Qzy,Qxz,Qyz,Qzz=M.flat
    # Fill only lower half of symmetric matrix
    K = np.array([
        [Qxx-Qyy-Qzz, 0, 0, 0],
        [Qyx+Qxy, Qyy-Qxx-Qzz, 0, 0],
        [Qzx+Qxz, Qzy+Qyz, Qzz-Qxx-Qyy, 0],
        [Qyz-Qzy, Qzx-Qxz, Qxy-Qyx, Qxx+Qyy+Qzz]]) / 3
    # Use Hermitian eigenvectors, values for speed
    vals, vecs = np.linalg.eigh(K)
    # Select largest eigenvector, reorder to w,x,y,z quaternion
    q = vecs[[3, 0, 1, 2],np.argmax(vals)]
    # Prefer quaternion with positive w
    # (q * -1 corresponds to same rotation as q)
    if q[0]<0:
        q *= -1
    return q


def mult(q1, q2):
    ''' Multiply two quaternions

    Parameters
    ----------
    q1 : 4 element sequence
    q2 : 4 element sequence

    Returns
    -------
    q12 : shape (4,) array

    Notes
    -----
    See : http://en.wikipedia.org/wiki/Quaternions#Hamilton_product
    '''
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1*w2 - x1*x2 - y1*y2 - z1*z2
    x = w1*x2 + x1*w2 + y1*z2 - z1*y2
    y = w1*y2 + y1*w2 + z1*x2 - x1*z2
    z = w1*z2 + z1*w2 + x1*y2 - y1*x2
    return np.array([w, x, y, z])


def conjugate(q):
    ''' Conjugate of quaternion 

    Parameters
    ----------
    q : 4 element sequence
       w, i, j, k of quaternion

    Returns
    -------
    conjq : array shape (4,)
       w, i, j, k of conjugate of `q`
    '''
    return np.array(q) * np.array([1.0,-1,-1,-1])


def norm(q):
    ''' Return norm of quaternion 

    Parameters
    ----------
    q : 4 element sequence
       w, i, j, k of quaternion

    Returns
    -------
    n : scalar
       quaternion norm
    '''    
    return np.dot(q,q)


def isunit(q):
    ''' Return True is this is very nearly a unit quaternion '''
    return np.allclose(norm(q),1)


def inverse(q):
    ''' Return multiplicative inverse of quaternion `q`

    Parameters
    ----------
    q : 4 element sequence
       w, i, j, k of quaternion

    Returns
    -------
    invq : array shape (4,)
       w, i, j, k of quaternion inverse
    '''
    return conjugate(q) / norm(q)


def eye():
    ''' Return identity quaternion '''
    return np.array([1.0,0,0,0])


def rotate_vector(v, q):
    ''' Apply transformation in quaternion `q` to vector `v`

    Parameters
    ----------
    v : 3 element sequence
       3 dimensional vector
    q : 4 element sequence
       w, i, j, k of quaternion

    Returns
    -------
    vdash : array shape (3,)
       `v` rotated by quaternion `q`

    Notes
    -----
    See: http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Describing_rotations_with_quaternions
    
    '''
    varr = np.zeros((4,))
    varr[1:] = v
    return mult(q, mult(varr, conjugate(q)))[1:]


def nearly_equivalent(q1, q2, rtol=1e-5, atol=1e-8):
    ''' Returns True if `q1` and `q2` give near equivalent transforms

    q1 may be nearly numerically equal to q2, or nearly equal to q2 * -1
    (becuase a quaternion multiplied by -1 gives the same transform).

    Parameters
    ----------
    q1 : 4 element sequence
       w, x, y, z of first quaternion
    q2 : 4 element sequence
       w, x, y, z of second quaternion
    
    Returns
    -------
    equiv : bool
       True if `q1` and `q2` are nearly equivalent, False otherwise

    Examples
    --------
    >>> q1 = [1, 0, 0, 0]
    >>> nearly_equivalent(q1, [0, 1, 0, 0])
    False
    >>> nearly_equivalent(q1, [1, 0, 0, 0])
    True
    >>> nearly_equivalent(q1, [-1, 0, 0, 0])
    True
    '''
    q1 = np.array(q1)
    q2 = np.array(q2)
    if np.allclose(q1, q2, rtol, atol):
        return True
    return np.allclose(q1 * -1, q2, rtol, atol)


def axangle2quat(vector, theta, is_normalized=False):
    ''' Quaternion for rotation of angle `theta` around `vector`

    Parameters
    ----------
    vector : 3 element sequence
       vector specifying axis for rotation.
    theta : scalar
       angle of rotation
    is_normalized : bool, optional
       True if vector is already normalized (has norm of 1).  Default
       False

    Returns
    -------
    quat : 4 element sequence of symbols
       quaternion giving specified rotation

    Examples
    --------
    >>> q = axangle2quat([1, 0, 0], np.pi)
    >>> np.allclose(q, [0, 1, 0,  0])
    True

    Notes
    -----
    Formula from http://mathworld.wolfram.com/EulerParameters.html
    '''
    vector = np.array(vector)
    if not is_normalized:
        # Not in place to avoid numpy's stricter casting rules
        vector = vector / math.sqrt(np.dot(vector, vector))
    t2 = theta / 2.0
    st2 = math.sin(t2)
    return np.concatenate(([math.cos(t2)],
                           vector * st2))


def axangle2rmat(vector, theta):
    ''' Rotation matrix of angle `theta` around `vector`

    Parameters
    ----------
    vector : 3 element sequence
       vector specifying axis for rotation.
    theta : scalar
       angle of rotation

    Returns
    -------
    mat : array shape (3,3)
       rotation matrix specified rotation

    Notes
    -----
    From: http://en.wikipedia.org/wiki/Rotation_matrix#Axis_and_angle
    '''
    x, y, z = vector[:3]
    n = math.sqrt(x*x + y*y + z*z)
    x = x/n
    y = y/n
    z = z/n
    c = math.cos(theta); s = math.sin(theta); C = 1-c
    xs = x*s;   ys = y*s;   zs = z*s
    xC = x*C;   yC = y*C;   zC = z*C
    xyC = x*yC; yzC = y*zC; zxC = z*xC
    return np.array([
            [ x*xC+c,   xyC-zs,   zxC+ys ],
            [ xyC+zs,   y*yC+c,   yzC-xs ],
            [ zxC-ys,   yzC+xs,   z*zC+c ]])


def quat2axangle(quat, identity_thresh=None):
    ''' Convert quaternion to rotation of angle around axis

    Parameters
    ----------
    quat : 4 element sequence
       w, x, y, z forming quaternion
    identity_thresh : None or scalar, optional
       threshold below which the norm of the vector part of the
       quaternion (x, y, z) is deemed to be 0, leading to the identity
       rotation.  None (the default) leads to a threshold estimated
       based on the precision of the input.
       
    Returns
    -------
    theta : scalar
       angle of rotation
    vector : array shape (3,)
       axis around which rotation occurs

    Examples
    --------
    >>> vec, theta = quat2axangle([0, 1, 0, 0])
    >>> vec
    array([ 1.,  0.,  0.])
    >>> np.allclose(theta, np.pi)
    True

    If this is an identity rotation, we return a zero angle and an
    arbitrary vector
    
    >>> quat2axangle([1, 0, 0, 0])
    (array([ 1.,  0.,  0.]), 0.0)

    Notes
    -----
    A quaternion for which x, y, z are all equal to 0, is an identity
    rotation.  In this case we return a 0 angle and an  arbitrary 
    vector, here [1, 0, 0]
    '''
    w, x, y, z = quat
    vec = np.asarray([x, y, z])
    if identity_thresh is None:
        try:
            identity_thresh = np.finfo(vec.dtype).eps * 3
        except ValueError:  # integer type
            identity_thresh = _FLOAT_EPS * 3
    len2 = x * x + y * y + z * z
    if len2 < identity_thresh ** 2:
        # if vec is nearly 0,0,0, this is an identity rotation
        return np.array([1.0, 0, 0]), 0.0
    theta = 2 * math.acos(max(min(w, 1), -1))
    if len2 == float('inf'):
        return np.zeros((3,)), theta
    return vec / math.sqrt(len2), theta



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
import os

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('statistics', parent_package, top_path)
    config.add_subpackage('models')
    config.add_subpackage('formula')
    config.add_subpackage('bench')
    config.add_subpackage('tests')
    config.add_include_dirs(config.name.replace('.', os.sep))
    config.add_extension('intvol', 'intvol.pyx',
                         include_dirs=[np.get_include()])
    config.add_extension('histogram', 'histogram.pyx',
                         include_dirs=[np.get_include()])
    config.add_extension('_quantile',
                         sources=['_quantile.pyx', 'quantile.c'])
    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())
